#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

"""

import numpy as np
from numpy import ma


def bin_spike(x, l):
    """

        l is the number of points used for comparison, thus l=2 means that each
          point will be compared only against the previous and following
          measurements. l=2 is is probably not a good choice, too small.

        Maybe use pstsd instead?

        Dummy way to avoid warnings when x[ini:fin] are all masked.
        Improve this in the future.
    """
    assert x.ndim == 1, "I'm not ready to deal with multidimensional x"

    assert l%2 == 0, "l must be an even integer"

    N = len(x)
    bin = ma.masked_all(N)
    # bin_std = ma.masked_all(N)
    half_window = int(l/2)
    idx = (i for i in range(half_window, N - half_window) if np.isfinite(x[i]))
    for i in idx:
        ini = max(0, i - half_window)
        fin = min(N, i + half_window)
        # At least 3 valid points
        if ma.compressed(x[ini:fin]).size >= 3:
            bin[i] = x[i] - ma.median(x[ini:fin])
            # bin_std[i] = (np.append(x[ini:i], x[i+1:fin+1])).std()
            bin[i] /= (np.append(x[ini:i], x[i+1:fin+1])).std()

    return bin


class Bin_Spike(object):
    def __init__(self, data, varname, cfg, autoflag=True):
        self.data = data
        self.varname = varname
        self.cfg = cfg

        self.set_features()
        if autoflag:
            self.test()

    def keys(self):
        return self.features.keys() + \
            ["flag_%s" % f for f in self.flags.keys()]

    def set_features(self):
        self.features = {'bin_spike': bin_spike(self.data[self.varname],
            self.cfg['l'])}

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg['threshold']
        except:
            print("Deprecated cfg format. It should contain a threshold item.")
            threshold = self.cfg

        try:
            flag_good = self.cfg['flag_good']
            flag_bad = self.cfg['flag_bad']
        except:
            print("Deprecated cfg format. It should contain flag_good & flag_bad.")
            flag_good = 1
            flag_bad = 3

        assert (np.size(threshold) == 1) and \
                (threshold is not None) and \
                (np.isfinite(threshold))   

        flag = np.zeros(self.data[self.varname].shape, dtype='i1')
        flag[np.nonzero(self.features['bin_spike'] > threshold)] = flag_bad
        flag[np.nonzero(self.features['bin_spike'] <= threshold)] = flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags['bin_spike'] = flag



import numpy as np

# ============================================================================
# I need to improve this, and include the places where the
#   flags are masked, i.e. only eliminate where the flags
#   could guarantee it was false.


def combined_flag(flags, criteria=None):
    """ Returns the combined flag considering all the criteria

        Input: flags

        Collects all flags in the criteria, and for each measurements, it
          return the maximum flag value among the different criteria.

        If criteria is not defined, considers all the flags,
          i.e. flags.keys()
    """
    assert hasattr(flags, 'keys')

    if criteria is None:
        criteria = list(flags.keys())

    output = np.asanyarray(flags[criteria[0]])
    for c in criteria[1:]:
        assert len(flags[c]) == len(output)
        output = np.max([output, flags[c]], axis=0)

    return output


def make_qc_index(flags, criteria, type="anytrue"):
    ind = flags[criteria[0]].copy()
    if type == "anytrue":
        for c in criteria:
            ind[(ind is True) | (flags[c] is True)] = True
        #ind[np.nonzero((ind == True) | (np.array(flags[c]) == True))[0]] = True
    elif type == "alltrue":
        for c in criteria:
            ind[(ind is True) | (flags[c] is True)] = True
    for c in criteria:
        ind[(ind is False) | (flags[c] is False)] = False
        # ind[np.nonzero((ind == False) | (np.array(flags[c]) == False))[0]] = False
    return ind



""" Apply Quality Control of CTD profiles
"""

import pkg_resources
from copy import deepcopy
from datetime import datetime
from os.path import basename
import re
import json
import logging

import numpy as np
from numpy import ma

from cotede.qctests import *
from cotede.misc import combined_flag
from cotede.utils import load_cfg


module_logger = logging.getLogger(__name__)


class ProfileQC(object):
    """Quality Control a CTD profile
    """

    def __init__(self, input, cfg=None, saveauxiliary=True, verbose=True,
            attributes=None):
        """A procedure to QC a hydrographic profile

        Parameters
        ----------
        input: dict-like
            An object with the data to be evaluated that responds like a
            dictionary. For instance, a variable pressure should be acessible
            as input['pressure'], or temperature as input['temperature'].
            This input object could have attrs, with global attributes for
            the whole dataset. For instance, input.attrs['lat'] would give the
            nominal latitude of the dataset input.

        cfg: dict-like or str
            The QC configuration to be used in the current profile. If a
            string, it should be the name of a JSON QC configuration. Check
            the manual for the available options.

        saveauxiliary: bool
            Save features as .features

        verbose: bool
            Show extra information

        attributes: dict-like, optional
            If given, append/overwirte the input.attrs

        Methods
        -------
        keys(self): List of input contents
        """
        # self.logger = logging.getLogger(logger or 'cotede.ProfileQC')

        try:
            self.name = input.filename
        except:
            self.name = None
        self.verbose = verbose

        if attributes is None:
            assert (hasattr(input, 'attributes'))
        assert (hasattr(input, 'keys')) and (len(input.keys()) > 0)

        self.cfg = load_cfg(cfg)
        module_logger.debug("Using cfg: {}".format(self.cfg))

        self.input = deepcopy(input)
        if attributes is None:
            self.attrs = input.attributes
        else:
            self.attrs = attributes
        self.flags = {}
        self.saveauxiliary = saveauxiliary
        if saveauxiliary:
            #self.auxiliary = {}
            # build_auxiliary is not exactly the best way to do it.
            self.build_features()

        if 'common' in self.cfg:
            self.evaluate_common(self.cfg)

        for v in self.input.keys():
            if v == 'TEMP':
                vv = 'sea_water_temperature'
            elif v == 'PSAL':
                vv = 'sea_water_salinity'
            else:
                vv = v
            for c in self.cfg['variables']:
                if re.match("(%s)2?$" % c, vv):
                    module_logger.debug(" %s - evaluating: %s, as type: %s" %
                                            (self.name, v, c))
                    self.evaluate(v, self.cfg['variables'][c])
                    break

    @property
    def data(self):
        return self.input.data

    @property
    def attributes(self):
        """Temporary solution while migrating attributes -> attrs
        """
        return self.attrs

    @property
    def auxiliary(self):
        module_logger.warning('ATENTION: Please use .features instead.'
                              'auxiliary will be eventually removed.')
        return self.features

    def keys(self):
        """ Return the available keys in self.data
        """
        return self.input.keys()

    def __getitem__(self, key):
        """ Return the key array from self.data
        """
        return self.input[key]

    def evaluate_common(self, cfg):
        self.flags['common'] = {}

        if 'valid_datetime' in self.cfg['common']:
            if 'datetime' in self.attrs.keys() and \
                    type(self.attrs['datetime']) == datetime:
                f = 1
            else:
                f = 3
            self.flags['common']['valid_datetime'] = f

        if 'datetime_range' in self.cfg['common']:
            if 'datetime' in self.attrs.keys() and \
                    (self.attrs['datetime'] >=
                            self.cfg['common']['datetime_range']['minval']) and \
                    (self.attrs['datetime'] <=
                            self.cfg['common']['datetime_range']['maxval']):
                f = 1
            else:
                f = 3
            self.flags['common']['datetime_range'] = f

        if 'location_at_sea' in self.cfg['common']:
            y = LocationAtSea(self.input, cfg['common']['location_at_sea'])

            if self.saveauxiliary:
                for f in y.features.keys():
                    self.features['common'][f] = y.features[f]
            for f in y.flags:
                self.flags['common'][f] = y.flags[f]

        # if self.saveauxiliary:
        #     self.features['common'] = {}
        #     # Need to improve this. descentPrate doesn't make sense
        #     #   for Argo. That's why the try.
        #     try:
        #         self.features['common']['descentPrate'] = \
        #                 descentPrate(self.input)
        #     except:
        #         pass

    def evaluate(self, v, cfg):

        self.flags[v] = {}

        # Apply common flag for all points.
        if 'common' in self.flags:
            N = self.input[v].shape
            for f in self.flags['common']:
                self.flags[v][f] = self.flags['common'][f] * \
                        np.ones(N, dtype='i1')

        if self.saveauxiliary:
            if v not in self.features.keys():
                self.features[v] = {}

        if 'platform_identification' in cfg:
            module_logger.warning(
                    "Sorry I'm not ready to evaluate platform_identification()")

        if 'valid_geolocation' in cfg:
            y = ValidGeolocation(
                    self.input, v, cfg['valid_geolocation'], autoflag=True)

            if self.saveauxiliary:
                for f in y.features.keys():
                    self.features[v][f] = y.features[f]
            for f in y.flags:
                self.flags[v][f] = y.flags[f]

        if 'valid_speed' in cfg:
            # Think about. Argo also has a test valid_speed, but that is
            #   in respect to sucessive profiles. How is the best way to
            #   distinguish them here?
            try:
                if self.saveauxiliary:
                    self.flags[v]['valid_speed'], \
                            self.features[v]['valid_speed'] = \
                            possible_speed(self.input, cfg['valid_speed'])
            except:
                module_logger.warning("Fail on valid_speed")

        if 'pressure_increasing' in cfg:
            module_logger.warning(
                    "Sorry, I'm no ready to evaluate pressure_increasing()")

        if 'grey_list' in cfg:
            module_logger.warning("Sorry I'm not ready to evaluate grey_list()")

        if 'gross_sensor_drift' in cfg:
            module_logger.warning(
                    "Sorry I'm not ready to evaluate gross_sensor_drift()")

        if 'frozen_profile' in cfg:
            module_logger.warning(
                    "Sorry I'm not ready to evaluate frozen_profile()")

        catalog = {
                'bin_spike': Bin_Spike,
                'cars_normbias': CARS_NormBias,
                'constant_cluster_size': ConstantClusterSize,
                'cum_rate_of_change': CumRateOfChange,
                'deepest_pressure': DeepestPressure,
                'digit_roll_over': DigitRollOver,
                'global_range': GlobalRange,
                'gradient': Gradient,
                'gradient_depthconditional': GradientDepthConditional,
                'monotonic_z': MonotonicZ,
                'profile_envelop': ProfileEnvelop,
                'rate_of_change': RateOfChange,
                'regional_range': RegionalRange,
                'spike': Spike,
                'spike_depthconditional': SpikeDepthConditional,
                'stuck_value': StuckValue,
                'tukey53H_norm': Tukey53H,
                'woa_normbias': WOA_NormBias,
                'density_inversion': DensityInversion,
                }

        for criterion in [c for c in catalog if c in cfg]:
            Procedure = catalog[criterion]
            try:
                y = Procedure(self.input, v, cfg[criterion], autoflag=True)
            except:
                # currently accommodates catalog['density_inversion']
                y = Procedure(self.input, cfg[criterion], autoflag=True)

            if self.saveauxiliary:
                for f in y.features.keys():
                    self.features[v][f] = y.features[f]
            for f in y.flags:
                self.flags[v][f] = y.flags[f]
        #if 'spike_depthsmooth' in cfg:
        #    from maud.window_func import _weight_hann as wfunc
        #    cfg_tmp = cfg['spike_depthsmooth']
        #    cfg_tmp['dzwindow'] = 10
        #    smooth = ma.masked_all(self.input[v].shape)
        #    z = ped['pressure']
        #    for i in range(len(self.input[v])):
        #        ind = np.nonzero(ma.absolute(z-z[i]) < \
        #                cfg_tmp['dzwindow'])[0]
        #        ind = ind[ind != i]
        #        w = wfunc(z[ind]-z[i], cfg_tmp['dzwindow'])
        #        smooth[i] = (T[ind]*w).sum()/w.sum()


        #if 'pstep' in cfg:
        #    ind = np.isfinite(self.input[v])
        #    ind = ma.getmaskarray(self.input[v])
        #    if self.saveauxiliary:
        #        self.features[v]['pstep'] = ma.concatenate(
        #                [ma.masked_all(1),
        #                    np.diff(self.input['PRES'][ind])])

        # FIXME: the Anomaly Detection and Fuzzy require some features
        #   to be estimated previously. Generalize this.
        if 'anomaly_detection' in  cfg:
            features = {}
            for f in cfg['anomaly_detection']['features']:
                try:
                    features[f] = self.features[v][f]
                except:
                    if f == 'spike':
                        features['spike'] = spike(self.input[v])
                    elif f == 'gradient':
                        features['gradient'] = gradient(self.input[v])
                    elif f == 'constant_cluster_size':
                        features['constant_cluster_size'] = \
                                constant_cluster_size(self.input[v])
                    elif f == 'tukey53H_norm':
                        features['tukey53H_norm'] = tukey53H_norm(self.input[v])
                    elif f == 'rate_of_change':
                        features['rate_of_change'] = rate_of_change(self.input[v])
                    elif (f == 'woa_normbias'):
                        y = WOA_NormBias(self.input, v, {}, autoflag=False)
                        features['woa_normbias'] = \
                                np.abs(y.features['woa_normbias'])
                    elif (f == 'cars_normbias'):
                        y = CARS_NormBias(self.input, v, {}, autoflag=False)
                        features['cars_normbias'] = \
                                np.abs(y.features['cars_normbias'])
                    else:
                        module_logger.error(
                                "Sorry, I can't evaluate anomaly_detection with: %s" % f)

            prob, self.flags[v]['anomaly_detection'] = \
                    anomaly_detection(features, cfg['anomaly_detection'])

            if self.saveauxiliary:
                self.features[v]['anomaly_detection'] = prob

        if 'morello2014' in cfg:
            self.flags[v]['morello2014'] = morello2014(
                    features=self.features[v],
                    cfg=cfg['morello2014'])

        if 'fuzzylogic' in  cfg:
            features = {}
            for f in cfg['fuzzylogic']['features']:
                try:
                    features[f] = self.features[v][f]
                except:
                    module_logger.error("Can't evaluate fuzzylogic with: %s" % f)

            self.flags[v]['fuzzylogic'] = fuzzylogic(
                    features=features,
                    cfg=cfg['fuzzylogic'])

        self.flags[v]['overall'] = combined_flag(self.flags[v])

    def build_features(self):
        if not hasattr(self, 'features'):
            self.features = {}

        self.features['common'] = {}
        # try:
        #     self.features['common']['descentPrate'] = \
        #             descentPrate(self.input)
        # except:
        #     logging.warn("Failled to run descentPrate")


class ProfileQCed(ProfileQC):
    """
    """
    def __init__(self, input, cfg=None):
        """
        """
        self.name = 'ProfileQCed'
        super(ProfileQCed, self).__init__(input, cfg)

    def keys(self):
        """ Return the available keys in self.data
        """
        return self.input.keys()

    def __getitem__(self, key):
        """ Return the key array from self.data
        """
        if key not in self.flags.keys():
            return self.input[key]
        else:
            return ma.masked_array(self.input[key].data,
                    mask=(self.flags[key]['overall']!=1))

        raise KeyError('%s not found' % key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

"""

import logging

import numpy as np
from numpy import ma

from cotede.qctests import QCCheckVar

try:
    import pandas as pd

    PANDAS_AVAILABLE = True
except:
    PANDAS_AVAILABLE = False

module_logger = logging.getLogger(__name__)


def gradient(x):
    return curvature(x)


def _curvature_pandas(x):
    """Equivalent to curvature() but using pandas

    It looks like the numpy implementation is faster even for larger datasets,
    so the default is with numpy.

    Note
    ----
    - In the future this will be useful to handle specific window widths.
    """
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    if not PANDAS_AVAILABLE:
        return _curvature_numpy(x)

    if hasattr(x, "to_series"):
        x = x.to_series()
    elif not isinstance(x, pd.Series):
        x = pd.Series(x)

    y = np.nan * x
    y = x - (x.shift(1) + x.shift(-1)) / 2.0
    return np.array(y)


def curvature(x):
    """Curvature of a timeseries

    This test is commonly known as gradient for historical reasons, but that
    is a bad name choice since it is not the actual gradient, like:
    d/dx + d/dy + d/dz,
    but as defined by GTSPP, EuroGOOS and others, which is actually the
    curvature of the timeseries..

    Note
    ----
    - Pandas.Series operates with indexes, so it should be done different. In
      that case, call for _curcature_pandas.
    """
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    if PANDAS_AVAILABLE and isinstance(x, pd.Series):
        return _curvature_pandas(x)

    y = np.nan * np.ones_like(x)
    y[1:-1] = x[1:-1] - (x[:-2] + x[2:]) / 2.0
    return y


class Gradient(QCCheckVar):
    def set_features(self):
        self.features = {"gradient": curvature(self.data[self.varname])}

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            module_logger.debug(
                "Deprecated cfg format. It should contain a threshold item."
            )
            threshold = self.cfg

        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = np.absolute(self.features["gradient"])
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["gradient"] = flag



# -*- coding: utf-8 -*-

""" Procedures for visual inspection and human flagging.
"""

import numpy as np
from numpy import ma

try:
    import pylab
    import matplotlib.pyplot as plt
except:
    print('matplotlib is not available')


class HumanQC(object):
    """ Plot a profile and collect human QC evaluation

        Still have much to improve here.
    """
    def __init__(self):
        pass

    def eval(self, x, z, baseflag=None, fails=None,
            humanflag=None, refname=None, clim=None):
        """
        """
        self.x = np.asanyarray(x)
        self.z = np.asanyarray(z)
        self.clim = clim

        assert x.shape == z.shape, "Data and z coordinate must have same shape"

        if baseflag is None:
            self.baseflag = ma.ones(x.size).astype('bool')
        else:
            self.baseflag = np.asanyarray(baseflag)
            assert self.baseflag.shape == self.x.shape

        if humanflag is None:
            self.humanflag = ma.masked_all(self.x.size,
                    dtype='object')
        else:
            self.humanflag = np.asanyarray(humanflag)
            assert self.baseflag.shape == self.x.shape

        if fails is None:
            self.fails = ma.zeros(x.size).astype('bool')
        else:
            self.fails = np.asanyarray(fails)
            assert self.fails.shape == self.x.shape

        self.refname = refname

        self.plot()
        return self.humanflag

    def plot(self):
        """
        """
        self.fig, self.ax = plt.subplots()
        try:
            self.ax.fill_betweenx(self.clim['z'],
                    self.clim['mn'] - 6*self.clim['std'],
                    self.clim['mn'] + 6*self.clim['std'],
                    color='r', alpha=0.2)
            self.ax.fill_betweenx(self.clim['z'],
                    self.clim['mn'] - 3*self.clim['std'],
                    self.clim['mn'] + 3*self.clim['std'],
                    color='g', alpha=0.2)
        except:
            print("Wasn't able to plot climatology")
        self.line, = self.ax.plot(self.x, self.z, 'b.',
                picker=10) # 5 points tolerance
        # Plot the bad ones
        self.ax.plot(self.x[self.baseflag==False],
                self.z[self.baseflag==False], 'r^')

        # Plot the dubious ones
        self.ax.plot(
                self.x[self.humanflag=='doubt'],
                self.z[self.humanflag=='doubt'],
                'D', color='magenta')

        self.ax.plot(self.x[self.fails], self.z[self.fails], 'o', ms=12,
                alpha=0.4, color='cyan', visible=True)


        #line, = ax.plot(xs, ys, 'o', picker=5)  # 5 points tolerance
        self.selected, = self.ax.plot([self.x[0]], [self.z[0]], 'o', ms=12,
                alpha=0.4, color='yellow', visible=False)

        self.hbad, = self.ax.plot([], [], '^', ms=12,
                alpha=0.4, color='red', visible=True)
        self.hgood, = self.ax.plot([], [], 's', ms=12,
                alpha=0.4, color='green', visible=True)
        self.hdoubt, = self.ax.plot([], [], 'D', ms=12,
                alpha=0.4, color='magenta', visible=True)

        self.ax.set_ylabel('Pressure')
        self.ax.set_ylim(self.ax.get_ylim()[::-1])

        self.fig.canvas.mpl_connect('close_event', self.handle_close)
        self.fig.canvas.mpl_connect('pick_event', self.onpick)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key)
        #pylab.connect('button_press_event', self.onpick)

        if self.refname is not None:
            self.ax.title(refname)

        pylab.show()


    def draw_humanflags(self, ind, flag):
        assert flag in [None, 'good', 'bad', 'doubt']

        if flag is None:
            self.humanflag.mask[self.dataind] = True
        else:
            self.humanflag[ind] = flag

        self.hbad.set_data(
                self.x[self.humanflag=='bad'],
                self.z[self.humanflag=='bad'])
        self.hgood.set_data(
                self.x[self.humanflag=='good'],
                self.z[self.humanflag=='good'])
        self.hdoubt.set_data(
                self.x[self.humanflag=='doubt'],
                self.z[self.humanflag=='doubt'])
        self.fig.canvas.draw()

    def on_key(self, event):
        if (event.key == 'r') :
            #self.fig.canvas.draw()
            plt.close()
            self.plot()

        elif (event.key == 'c') :
            print("Removing %s from the human list" % self.dataind)
            self.draw_humanflags(self.dataind, None)

        elif (event.key == 'f') :
            print("Adding %s on the bad list" % self.dataind)
            self.draw_humanflags(self.dataind, 'bad')

        elif (event.key == 't') :
            print("Adding %s on the good list" % self.dataind)
            self.draw_humanflags(self.dataind, 'good')

        elif (event.key == 'd') :
            print("Adding %s on the dubious list" % self.dataind)
            self.draw_humanflags(self.dataind, 'doubt')

        elif (event.key == 'z') :
            (xini, xfin, zini, zfin) = self.ax.axis()
            xrange = xfin - xini
            yrange = zfin - zini
            #pylab.axis([xini, xfin, zini, zfin])
            self.ax.set_xlim(self.x[self.dataind] - xrange/2., 
                    self.x[self.dataind] + xrange/2.)
            self.ax.set_ylim(self.z[self.dataind] - yrange/2., 
                    self.z[self.dataind] + yrange/2.)
            self.fig.canvas.draw()

        elif (event.key == 'q') :
            print("Quiting.")
            pylab.close()

    def onpick(self, event):
        """
        """
        if event.artist != self.line: return True

        N = len(event.ind)
        if not N: return True

        if N == 1:
            self.dataind = event.ind
        else:
            # Will get the closest point
            x = event.mouseevent.xdata
            y = event.mouseevent.ydata

            distances = np.hypot(x-self.x[event.ind], y-self.z[event.ind])
            indmin = distances.argmin()
            self.dataind = event.ind[indmin]

        self.selected.set_visible(True)
        self.selected.set_data(self.x[self.dataind], self.z[self.dataind])
        self.fig.canvas.draw()

    def handle_close(self, event):
        print('Closed Figure!')
        print("Good list: %s" % np.nonzero(self.humanflag=='good')[0])
        print("Bad list: %s" % np.nonzero(self.humanflag=='bad')[0])
        print("Doubt list: %s" % np.nonzero(self.humanflag=='doubt')[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)


def valid_geolocation(lat, lon):
    if np.shape(lat) != np.shape(lon):
        return False

    idx = (
        ma.greater_equal(lat, -90)
        & ma.less_equal(lat, 90)
        & ma.greater_equal(lon, -180)
        & ma.less_equal(lon, 360)
    )
    return idx


class ValidGeolocation(QCCheckVar):
    def test(self):
        self.flags = {}

        if ("LATITUDE" in self.data.keys()) and ("LONGITUDE" in self.data.keys()):
            lat = self.data["LATITUDE"]
            lon = self.data["LONGITUDE"]
        elif ("LATITUDE" in self.data.attrs) and ("LONGITUDE" in self.data.attrs):
            lat = self.data.attrs["LATITUDE"]
            lon = self.data.attrs["LONGITUDE"]
        else:
            self.flags["valid_position"] = self.flag_bad
            return

        idx = valid_geolocation(lat, lon)
        flag = np.zeros(np.shape(idx), dtype="i1")
        flag[~idx] = self.flag_bad
        flag[idx] = self.flag_good
        flag[ma.getmaskarray(idx)] = 9
        self.flags["valid_geolocation"] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

   Argo, test #12. (10C, 5PSU)

"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar
from .rate_of_change import rate_of_change


module_logger = logging.getLogger(__name__)


class DigitRollOver(QCCheckVar):
    def set_features(self):
        self.features = {"rate_of_change": rate_of_change(self.data[self.varname])}

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            module_logger.debug(
                "Deprecated cfg format. It should contain a threshold item."
            )
            threshold = self.cfg

        try:
            flag_good = self.cfg["flag_good"]
        except KeyError:
            flag_good = 1
        try:
            flag_bad = self.cfg["flag_bad"]
        except KeyError:
            flag_bad = 4

        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = ma.absolute(self.features["rate_of_change"])
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags["digit_roll_over"] = flag



# -*- coding: utf-8 -*-

"""

    Where is this test from? GTSPP?
"""
import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)


def profile_envelop(data, cfg, varname):
    """

        Probably not the best way to do this, but works for now.
    """
    # assert varname in data.keys()

    z = data["PRES"]
    x = data[varname]

    flag = np.zeros(z.shape, dtype="i1")

    for layer in cfg["layers"]:
        ind = np.nonzero(eval("(z %s) & (z %s)" % (layer[0], layer[1])))[0]
        f = eval("(x[ind] > %s) & (x[ind] < %s)" % (layer[2], layer[3]))

        flag[ind[f == True]] = 1
        flag[ind[f == False]] = 4

    flag[ma.getmaskarray(x)] = 9

    return flag


class ProfileEnvelop(QCCheckVar):
    def test(self):
        self.flags = {}

        z = self.data["PRES"]
        x = self.data[self.varname]

        assert z.shape == x.shape

        flag = np.zeros(x.shape, dtype="i1")

        assert "layers" in self.cfg, "Profile envelop cfg requires layers"

        for layer in self.cfg['layers']:
            ind = np.nonzero(eval("(z %s) & (z %s)" % (layer[0], layer[1])))[0]
            f = eval("(x[ind] > %s) & (x[ind] < %s)" % (layer[2], layer[3]))

            flag[ind[f == True]] = self.flag_good
            flag[ind[f == False]] = self.flag_bad

        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags["profile_envelop"] = flag

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

    Based on Timms 2011, p9597

    RoC(t) = x(t) - x(t-1)

"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar


module_logger = logging.getLogger(__name__)


def rate_of_change(x):
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    y = x * np.nan
    y[1:] = np.diff(x)

    return y


class RateOfChange(QCCheckVar):
    def set_features(self):
        self.features = {"rate_of_change": rate_of_change(self.data[self.varname])}

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            print("Deprecated cfg format. It should contain a threshold item.")
            threshold = self.cfg

        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )

        feature = np.absolute(self.features["rate_of_change"])
        if ("sd_scale" in self.cfg) and self.cfg["sd_scale"]:
            feature /= feature.std()

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["rate_of_change"] = flag



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst


import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)


class StuckValue(QCCheckVar):
    def test(self):
        self.flags = {}

        x = ma.compressed(self.data[self.varname])
        flag = np.zeros(self.data[self.varname].shape, dtype='i1')
        if (x.size > 1) and (np.allclose(x, np.ones_like(x) * x[0])):
            flag[:] = self.flag_bad
        else:
            flag[:] = self.flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags['stuck_value'] = flag



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst


"""

    Based on Timms 2011, p9597

"The first case in Equation (1) ensures that the cumulative rate of change function is able to quickly respond to degradations in sensor quality, while the second case ensures that it takes a longer time for confidence to return in the measurements from that sensor once the rate of change of the parameter decreases." (Timms 2011)

        Small   S->M            Medium          M->L            Large
Temp    <0.03   0.03 to 0.05    0.05 to 0.07    0.07 to 0.11    >0.11
Cond    <50     50 to 100       100 to 150      150 to 250      >250


Timms 2011 uses k=0.8

if (RoC_small (t) + 0.5 × RoC_medium (t)) < (cRoC_small (t - 1) + 0.5 × cRoC_medium (t - 1)):
    cRoC_i (t) = RoC_i (t)
else
    cRoC_i (t) = (1 - k) × RoC_i(t) + k * cRoC_i(t - 1)

i = small, medium, large
"""

import numpy as np
from numpy import ma
import logging

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)

def cum_rate_of_change(x, memory):
    """Cummulative rate of change
    """
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    y = np.nan * np.ones_like(x)
    y[1:] = np.absolute(np.diff(x))

    for i in range(2, y.size):
        if y[i] < y[i - 1]:
            y[i] = (1 - memory) * y[i] + memory * y[i - 1]

    return y


class CumRateOfChange(QCCheckVar):
    def set_features(self):
        module_logger.debug("Feature: cummulative rate of change")
        self.features = {
            "cum_rate_of_change": cum_rate_of_change(
                self.data[self.varname], self.cfg["memory"]
            )
        }

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            print("Deprecated cfg format. It should contain a threshold item.")
            threshold = self.cfg

        assert np.size(threshold) == 1, "Threshold should be a single value"
        assert threshold is not None, "Threshold can't be None"
        assert np.isfinite(threshold), "Threshold must be a valid number"

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = np.absolute(self.features["cum_rate_of_change"])
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["cum_rate_of_change"] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)
try:
    import shapely.wkt
    from shapely.geometry import Point
except ImportError:
    module_logger.debug("Module Shapely is not available")


class RegionalRange(QCCheckVar):
    """
        Two ways, define region with a wkt polygon or define the vertices of a simple rectangle

        Example of config:

        [{"name": "red_sea",
          "region": "POLYGON ((10, 40, 20, 50, 30, 30, 10, 40))",
          "minval": 21.7,
          "maxval": 40
          }]

    """

    def test(self):
        self.flags = {}

        feature = self.data[self.varname]

        try:
            import shapely
        except ImportError:
            module_logger.debug(
                "Regional range currently depends on module Shapely, which is not available. Regional range will return flag 0."
            )
            self.flags["regional_range"] = np.zeros(feature.shape, dtype="i1")
            return

        if ("LATITUDE" in self.data.keys()) and ("LONGITUDE" in self.data.keys()):
            lat = self.data["LATITUDE"]
            lon = self.data["LONGITUDE"]

            module_logger.warning("I'm not ready to handle regional range of a track")
            return
        elif ("LATITUDE" in self.data.attrs) and ("LONGITUDE" in self.data.attrs):
            lat = self.data.attrs["LATITUDE"]
            lon = self.data.attrs["LONGITUDE"]
        else:
            self.flags["regional_range"] = np.zeros(feature.shape, dtype="i1")
            return

        assert "regions" in self.cfg

        flag = np.zeros(feature.shape, dtype="i1")
        for cfg in self.cfg['regions']:

            assert "name" in cfg, "Regional Range must have a name"

            assert ("minval" in cfg) and (
                "maxval" in cfg
            ), "Missing limits: minval & maxval"

            minval = cfg["minval"]
            maxval = cfg["maxval"]

            assert minval < maxval, (
                "Regional Range(%s): "
                + "minval (%s) must be smaller than maxval(%s)" % (v, minval, maxval)
            )

            g = shapely.wkt.loads(cfg["region"])
            if g.intersects(Point(lon, lat)):
                flag[np.nonzero(feature < minval)] = self.flag_bad
                flag[np.nonzero(feature > maxval)] = self.flag_bad
                idx = (feature >= minval) & (feature <= maxval)
                idx = idx & (flag < self.flag_good)
                flag[np.nonzero(idx)] = self.flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags["regional_range"] = flag



# -*- coding: utf-8 -*-

"""
"""

import logging

from numpy import ma


module_logger = logging.getLogger(__name__)

def descentPrate(data):
    """

        It's probably a good idea to smooth it with a window of 2-5 seconds.
        After binned, the data will be probably groupped in bins of 1dbar,
          but the raw data might have more than one records per second, which
          might have plenty spikes. I'm looking here for inadequate casts
          lowered too fast, or maybe bad weather and a rolling vessel.

        Consider to create another test looking for excessive ups and downs.
    """
    assert ('timeS' in data.keys()), "timeS is not available"
    assert ('PRES' in data.keys()), "pressure is not available"
    assert data['timeS'].shape == data['PRES'].shape, \
            "t and p have different sizes"
    dt = ma.diff(data['timeS'])
    dp = ma.diff(data['PRES'])
    y = ma.append(ma.fix_invalid(np.nan), dp/dt)
    return y



# -*- coding: utf-8 -*-

"""
    Quality Control based on anomaly detection
"""

import logging

import numpy as np
#from numpy import ma
from cotede.anomaly_detection import estimate_anomaly


module_logger = logging.getLogger(__name__)

def anomaly_detection(features, cfg):
    """

        Must decide where to set the flags.
    """

    prob = estimate_anomaly(features, params = cfg['features'])
    #flag = np.zeros(self.input[v].shape, dtype='i1')
    flag = np.zeros(prob.shape, dtype='i1')

    flag[np.nonzero(prob >= cfg['threshold'])] = 1
    flag[np.nonzero(prob < cfg['threshold'])] = 4

    return prob, flag



# -*- coding: utf-8 -*-

"""

    This was initially inspired in the sequel: Timms 2011, Morello 2011,
      Morello 2014. I reproduced their work and also implemented an
      alternative approach using a more traditional defuzzification
      procedure.

"""

import numpy as np
from numpy import ma

from .membership_functions import smf, zmf, trapmf, trimf
from .defuzz import defuzz

def fuzzyfy(features, cfg):
    """

        FIXME: Looks like skfuzzy.trapmf does not handle well masked values.
               I must think better what to do with masked input values. What
               to do when there is one feature, but the other features are
               masked?
    """

    features_list = list(cfg['features'].keys())

    N = features[features_list[0]].size

    # The fuzzy set are usually: low, medium, high
    # The membership of each fuzzy set are each feature scaled.
    membership = {}
    for f in cfg['output'].keys():
        membership[f] = {}

    for t in features_list:
        for m in membership:
            assert m in cfg['features'][t], \
                    "Missing %s in %s" % (m, cfg['features'][t])

            membership[m][t] = ma.masked_all_like(features[t])
            ind = ~ma.getmaskarray(features[t])
            if m == 'low':
                membership[m][t][ind] = zmf(
                        np.asanyarray(features[t])[ind], cfg['features'][t][m])
            elif m == 'high':
                membership[m][t][ind] = smf(
                        np.asanyarray(features[t])[ind],
                        cfg['features'][t][m])
            else:
                membership[m][t][ind] = trapmf(
                        np.asanyarray(features[t])[ind],
                        cfg['features'][t][m])

    # Rule Set
    rules = {}
    # Low: u_low = mean(S_l(spike), S_l(clim)...)
    #u_low = np.mean([weights['spike']['low'],
    #    weights['woa_relbias']['low']], axis=0)

    tmp = membership['low'][features_list[0]]
    for f in features_list[1:]:
        tmp = ma.vstack((tmp, membership['low'][f]))

    # FIXME: If there is only one feature, it will return 1 value
    #          instead of an array with N values.
    rules['low'] = ma.mean(tmp, axis=0)

    # IMPROVE IT: Morello2014 doesn't even use the medium uncertainty,
    #   so no reason to estimate it. In the generalize this once the
    #   membership combining rules are defined in the cfg, so I can
    #   decide to use mean or max.
    if 'medium' in membership:
        # Medium: u_medium = mean(S_l(spike), S_l(clim)...)
        #u_medium = np.mean([weights['spike']['medium'],
        #    weights['woa_relbias']['medium']], axis=0)

        tmp = membership['medium'][features_list[0]]
        for f in features_list[1:]:
            tmp = ma.vstack((tmp, membership['medium'][f]))

        rules['medium'] = ma.mean(tmp, axis=0)

    # High: u_high = max(S_l(spike), S_l(clim)...)
    #u_high = np.max([weights['spike']['high'],
    #    weights['woa_relbias']['high']], axis=0)

    tmp = membership['high'][features_list[0]]
    for f in features_list[1:]:
        tmp = ma.vstack((tmp, membership['high'][f]))

    rules['high'] = ma.max(tmp, axis=0)

    return rules


def fuzzy_uncertainty(features, cfg):
    """

        Temporary solution. Under-development.
    """
    # It's not clear at Morello 2014 what is the operator K()
    # Q is the uncertainty, hence Q_low is the low uncertainty
    # Seems like K() is just a linear factor, which would give the level of uncertainty, like 0.1 for low, 0.5 for medium and 0.9 for high would define weights for each level?! I'm not sure. But the result would be a composite curve, so when the Qs are joinned it would give a curve with the possible values on Q (maybe multiple dimensions) and the y would be the composite result [0, 1].
    # Q_low = 0.1 * u_low   # K_low(u_low)
    # Q_medium = 0.5 * u_medium   # K_medium(u_medium)
    # Q_high = 0.9 *u_high   # K_high(u_high)

    # Bisector

    # They refer to Q_i x_l, which I understand as the uncertainty for each value for each level
    # It looks like the uncertainties of all tests of the three levels are groupped and ordered, and the bisector would be the value that would define the half of the area.
    # Is it x the observed value of hypotetical values?

    # CQ = bisector(Qs, ...

    rules = fuzzyfy(features, cfg)

    N_out = 100
    output_range = np.linspace(0, 1, N_out)
    output = {}
    output['low'] = trimf(output_range, cfg['output']['low'])
    if 'medium' in cfg['output']:
        output['medium'] = trimf(output_range, cfg['output']['medium'])
    output['high'] = smf(output_range, cfg['output']['high'])

    # FIXME: As it is now, it will have no zero flag value. Think about cases
    #   where some values in a profile would not be estimated, hence flag=0
    #   I think skfuzzy does not accept masked arrays?!?! That would be the
    #   limiting factor.

    N = rules[list(rules.keys())[0]].size
    # This would be the classic fuzzy approach.
    uncertainty = ma.masked_all(N)
    for i in range(N):
        aggregated = np.zeros(N_out)
        for m in rules:
            if rules[m][i] is not ma.masked:
                aggregated = np.fmax(aggregated,
                                     np.fmin(rules[m][i], output[m]))
        if aggregated.sum() > 0:
            uncertainty[i] = defuzz(output_range, aggregated, 'bisector')

    return uncertainty



# -*- coding: utf-8 -*-

"""


Target: Initially two main functionalities

- Quality control flagging
    - Input:
        - Lista de arquivo | diretório
        - Q.C. config (can be a file)
        - Features (just varnames)
    - Pandas Collection:
        - Load all files and return flags plus features (aux)
    - Split groups: fit, test, eval
    - Calibrate:
        - Fit PDF parameters for each feature using fit group
        - Define the optimal thresholds comparing with flags from Q.C. config
            - There is the possibility to human interaction to overwrite the flags from the auto Q.C.
        - Estimate the error
        - The threshold between good or bad flags were defined on the second step, but now consider the whole dataset. The highest probability of any bad value is the thresholds between 1 and 2. In other words, all data with higher probability that this threshold where good. Between this threshold and the optimal previously defined there were bad data, but it was mostly good ones, hence flag 2 which means probably good. With the same concept define the threshold between flags 3 and 4.

    With the coeficients determined, an independent procedure to flag all data. Independent because the anomaly detection flagging itself can simply be loaded by previsouly defined coeficients, hence start already from this point.
    - Fit features, on full DB, there is no split data
    - With the parameters on the previous step estimate the probability of each measurement
    - Create a list sorted by:
        - Produtorium of all probabilities or,
        - Min(P(x_i)), the lowest probability for each measurement
    - The output would be a list to feed the Human Q.C. system
"""

import numpy as np
from numpy import ma
# from scipy.stats import norm, rayleigh, expon, halfnorm, exponpow, exponweib
from scipy.stats import exponweib
# from scipy.stats import kstest

from cotede.misc import combined_flag
from cotede.humanqc import HumanQC


def fit_tests(features, q=0.90, verbose=False):
    """

        Input:
          features: a dictionary like with the numerical results from the
              QC tests. For example, the gradient test values, not the
              flags, but the floats itself, like
              {'gradient': ma.array([.23, .12, .08]), 'spike': ...}
              It also works with a pandas.DataFrame()

          q: The lowest percentile to be considered. For example, .90
              means that only the top 10% data (i.e. percentiles higher
              than .90) are considered in the fitting.
    """
    assert (q >= 0) & (q < 1), "q must be in [0, 1)"

    output = {}
    for f in features:
        # Sample only valid values
        samp = ma.compressed(features[f][np.isfinite(features[f])])
        # Identify the percentile q
        qlimit = np.percentile(samp, 1e2*q)
        # Restricts to the top q values
        samp = samp[samp > qlimit]
        if samp.any():
            param = exponweib.fit(samp)
            output[f] = {'param': param,
                    'qlimit': qlimit}

        if verbose is True:
            import pylab
            x = np.linspace(samp.min(), samp.max(), 100)
            pdf_fitted = exponweib.pdf(x, *param[:-2], loc=param[-2], scale=param[-1])
            pylab.plot(x, pdf_fitted, 'b-')
            pylab.hist(ma.array(samp), 100, normed=1, alpha=.3)
            pylab.title(f)
            pylab.show()

    return output


def estimate_anomaly(features, params, method='produtorium'):
    """ Estimate probability from PDF defined by params

        The output is the natural logarithm of the estimated probability.

        params are the parameters that define the PDF for each feature
          in features. This function estimate the combined probability of
          each row in features as the produtorium between the probabilities
          of the different features on the same row.

        ATENTION!! I should think more about what would I like from this
          function. What should happens in case of a masked feature? And
          if all features for one measurement are masked? Right now it
          simply don't add for the estimate, so that all features masked
          would lead to an expectation of 100% it's good.
    """
    assert hasattr(params, 'keys')
    assert hasattr(features, 'keys')

    features_names = list(features.keys())
    for k in params.keys():
        assert k in features_names, "features doesn't have: %s" % k

    prob = ma.masked_all(np.shape(features[features_names[0]]), dtype='f8')

    for t in params.keys():
        param = params[t]['param']
        valid = ~ma.fix_invalid(features[t]).mask

        tmp = exponweib.sf(np.asanyarray(features[t]),
                *param[:-2], loc=param[-2], scale=param[-1])
        # Arbitrary solution. No value can have a probability of 0.
        tmp[tmp == 0] = 1e-25
        p = ma.log(tmp)

        # If both are valid, operate as choosed method.
        ind = ~prob.mask & valid
        if method == 'produtorium':
            prob[ind] = prob[ind] + p[ind]
        elif method == 'min':
            prob[ind] = min(prob[ind], p[ind])
        else:
            assert "Invalid method: %s" % method

        # Update prob if new value is valid and prob is masked
        # Operate twice the first feature if moved above.
        ind = prob.mask & valid
        prob[ind] = p[ind]

    return prob


def estimate_p_optimal(prob, binflag, verbose=False):
    """ ATENTION: I'm not happy with this. Improve it

        Maybe use flag as input, and optimize to give 3 thresholds
    """
    assert prob.shape == binflag.shape
    assert binflag.dtype == 'bool'

    err = []
    p_limit = prob[np.nonzero(binflag)].min() - 0.1
    P = -np.arange(0, -p_limit, 0.1)
    N = P.size
    err = np.empty(N)
    false_negative = np.empty(N)
    false_positive = np.empty(N)
    for i, p in enumerate(P):
        # The nonzero is necessary in case binflag is a masked array.
        false_negative[i] = np.nonzero(prob[np.nonzero(binflag)] < p)[0].size
        false_positive[i] = np.nonzero(prob[np.nonzero(~binflag)] > p)[0].size
        err[i] = false_negative[i] + false_positive[i]

    if verbose is True:
        import pylab
        pylab.plot(P, err , 'b'); pylab.show()

    return P[err.argmin()], float(err.min())/prob.size#, {'P': P, 'err': err}


def calibrate4flags(flags, features, q=0.90, verbose=False):
    """ Adjust coeficients for Anomaly Detection to best reproduce given flags

        Inputs:
            flag_ref: Reference index. What the Anomaly Detection will try
                   to reproduce. Uses the True and Falses from flag_ref
                   to partition the data to be used to fit, to adjust
                   and to estimate the error.
            qctests: The tests used by the Anomaly Detection. One curve will
                   be fit for each test.
            aux: The auxiliary tests results from the ProfileQCCollection. It
                   is expected that the qctests are present in aux.
            q: The top q extreme tests results to be used on Anom. Detect.
                 For example q=0 will use all the data, while q=0.9 (default)
                 will use the percentile of 0.9, i.e. the top 10% values.

            Output: Returns a dictionary with
                err:
                err_ratio:
                false_negative:
                false_positive:
                p_optimal:
                params:

            Use the functions:
                split_data_groups()
                fit_tests()
                estimate_anomaly()
                estimate_p_optimal()

    """
    assert hasattr(features, 'keys')

    binflags = i2b_flags(flags)
    assert len(features[features.keys()[0]]) == len(binflags)

    indices = split_data_groups(binflags)
    params = fit_tests(features[indices['fit']], q=q)
    prob = estimate_anomaly(features, params)

    if verbose is True:
        pylab.hist(prob)
        pylab.show()

    p_optimal, test_err = estimate_p_optimal(prob[indices['test']],
            binflags[indices['test']])

    # Guarantee the the false_* indices will be np.array
    false_negative = (prob < p_optimal) & binflags
    false_negative[ma.getmaskarray(false_negative)] = False
    false_negative = np.array(false_negative)
    false_positive = (prob > p_optimal) & ~binflags
    false_positive[ma.getmaskarray(false_positive)] = False
    false_positive = np.array(false_positive)

    mistake = false_positive | false_negative

    # I can extract only .data, since split_data_groups already eliminated
    #   all non valid positions.
    #err = np.nonzero(false_negative)[0].size + \
    #        np.nonzero(false_positive)[0].size
    tot_misfit = np.nonzero(mistake)[0].size
    n_err = float(np.nonzero(mistake[indices['err']])[0].shape[0])
    #err_ratio = float(err)/prob[indices['ind_err']].size
    err_ratio = n_err/indices['err'].astype('i').sum()
    #false_negative = (prob < p_optimal) & \
    #    (flag_ref.data is True) & (ma.getmaskarray(flag_ref) is False)
    #false_positive = (prob > p_optimal) & \
    #    (flag_ref.data is False) & (ma.getmaskarray(flag_ref) is False)

    output = {'false_negative': false_negative,
            'false_positive': false_positive,
            'prob': prob,
            'p_optimal': p_optimal,
            'tot_misfit': tot_misfit,
            'n_err': n_err,
            'err_ratio': err_ratio,
            'params': params}

    return output


def split_data_groups(flag):
    """ Splits randomly the indices into fit, test and error groups

        Return a dictionary with 3 indices set:
            - ind_fit with 60% of the good
            - ind_test with 20% of the good and 50% of the bad
            - ind_eval with 20% of the good and 50% of the bad
    """
    assert flag.dtype == 'bool'

    N = flag.size
    ind_base = np.zeros(N) == 1
    ind_valid = ~ma.getmaskarray(flag)

    # ==== Good data ==================
    ind_good = np.nonzero((flag == True)  & ind_valid)[0]
    N_good = ind_good.size
    perm = np.random.permutation(N_good)
    N_test = int(round(N_good*.2))
    ind_test = ind_base.copy()
    ind_test[ind_good[perm[:N_test]]] = True
    ind_err = ind_base.copy()
    ind_err[ind_good[perm[N_test:2*N_test]]] = True
    ind_fit = ind_base.copy()
    ind_fit[ind_good[perm[2*N_test:]]] = True

    # ==== Bad data ===================
    ind_bad = np.nonzero((flag == False) & ind_valid)[0]
    N_bad = ind_bad.size
    perm = np.random.permutation(N_bad)
    N_test = int(round(N_bad*.5))
    ind_test[ind_bad[perm[:N_test]]] = True
    ind_err[ind_bad[perm[N_test:]]] = True

    return {'fit': ind_fit, 'test': ind_test, 'err': ind_err}


def i2b_flags(flags, good_flags=[1,2], bad_flags=[3,4]):
    """ Converts int flags (like IOC) into binary (T|F)

        If given a dictionary of flags, it will evaluate each item
          of the dictionary, and return:
          - True if all available values are True
          - False if any of the available values is False
          - Masked is all values are masked
    """

    if (hasattr(flags, 'keys')) and (np.ndim(flags) > 1):
        output= []
        for f in flags:
            output.append(i2b_flags(flags[f], good_flags, bad_flags))

        return ma.array(output).all(axis=0)

    flags = np.asanyarray(flags)
    assert flags.dtype != 'bool', "Input flags should not be binary"
    output = ma.masked_all(np.shape(flags), dtype='bool')

    for f in good_flags:
        output[flags == f] = True
    for f in bad_flags:
        output[flags == f] = False

    return output


def calibrate_anomaly_detection(datadir, varname, cfg=None):
    """ Calibrate coefficientes for Anomaly Detection

        Input:
            datadir: Directory with the data to be used on calibration
            varname: Variable to calibrate. For example: TEMP
            cfg: CoTeDe's QC configuration. Can be None for CoTeDe's default
                a name for one of the preset configuration files, or a dict

        Output:
            false_negative:
            false_positive:
            prob:
            p_optimal:
            err:
            err_ratio:
            params:

        Loads all the data in datadir, apply the Q.C. procedures according to
          cfg, and than calibrate params and p_optimal so that anomaly
          detection reproduces the combined flags from the Q.C.
    """
    import pandas as pd

    assert type(varname) is str, "varname must be a string"

    db = ProfilesQCPandasCollection(datadir, cfg=cfg, saveauxiliary=True)

    assert varname in db.keys(), "db does not contain variable %s" % varname

    # # Remove the value out of the possible range.
    # ind_outofrange = np.nonzero(db.flags[varname]['global_range'] != 1)
    # binflag.mask[ind_outofrange] = True
    # hardlimit_flags = ['global_range']
    ind = db.flags[varname]['global_range'] == 1
    #aux = db.auxiliary[varname][ind]
    #features = aux.drop(['id','profileid'], axis=1)
    features = db.auxiliary[varname][ind]
    #flags = db.flags[varname][ind].drop(['id','profileid', 'density_inversion'], axis=1)
    #flags = db.flags[varname][ind].drop(['density_inversion'], axis=1)
    #flags = combined_flag(flags)
    #flags = combined_flag(db.flags[varname][ind])
    #binflags = i2b_flags(flags)

    result = calibrate4flags(db.flags[varname][ind],
            db.auxiliary[varname][ind], q=0.90, verbose=False)

    #ind = ma.masked_all(len(flags), dtype='bool')
    #ind[(flags == 1) | (flags == 2)] = True
    #ind[(flags == 3) | (flags == 4)] = False
    #indices = split_data_groups(ind)
    #indices = split_data_groups(flags)


    #params = fit_tests(features[indices['fit']], q=.9)
    #prob = estimate_anomaly(features, params)

    #binflag = i2b_flagsflag(db.flags[varname][ind], reference_flags)

    #p_optimal, test_err = estimate_p_optimal(prob[indices['test']],
    #        i2b_flags(flags[indices['test']]))

    #false_negative = prob[indices['err'] & binflags] < p_optimal
    #false_positive = prob[indices['err'] & ~binflags] < p_optimal
    #err = np.nonzero(false_negative)[0].size + \
    #        np.nonzero(false_positive)[0].size
    #err_ratio = float(err)/prob[indices['err']].size

    #output = {'false_negative': false_negative,
    #        'false_positive': false_positive,
    #        'prob': prob,
    #        'p_optimal': p_optimal,
    #        'err': err,
    #        'err_ratio': err_ratio,
    #        'params': params}

    #result = adjust_anomaly_coefficients(binflag, qctests, aux)

    #return output
    return result


def human_calibrate_mistakes(data, varname, flagname, featuresnames, niter=5):
    """
    """
    q = 0.90
    assert varname in data

    import pandas as pd
    #data['id'] = range(data.shape[0])
    #data.set_index('id', drop=True, inplace=True)
    # Guarantee that the indices are unique.
    assert data.index.unique().shape[0] == data.shape[0]

    if 'human_flag' not in data:
        data['human_flag'] = ma.masked_all(data[flagname].shape,
                dtype='object')
    data['flag_calibrating'] = data[flagname].copy()
    data.loc[data.human_flag == 'good', 'flag_calibrating'] = 1
    data.loc[data.human_flag == 'bad', 'flag_calibrating'] = 4

    result = calibrate4flags(data['flag_calibrating'], data[featuresnames], q=q)

    error_log = [{'err': result['n_err'],
        'err_ratio': result['err_ratio'],
        'p_optimal': result['p_optimal']}]

    for i in range(niter):
        for v in ['false_positive', 'false_negative', 'prob']:
            data[v] = result[v]
        # Failures from AD to reproduce flags
        data['mistake'] = data['human_flag'].isnull() & \
                (data['false_positive'] | data['false_negative'])

        # AD's severity error is given by how far away was the estimated
        #   probability from the prob. threshold.
        data['derr'] = (data.prob - result['p_optimal']).abs()
        data.loc[data.mistake == False, 'derr'] = np.nan

        grp = data[data.mistake].groupby('profileid')
        profileids = grp['derr'].max().sort_values(ascending=False).index

        # In the future order by how badly AD mistaked
        if len(profileids) == 0:
            break
        # 5 random profiles with mistakes
        # for pid in profileids[:10]:
        for pid in np.random.permutation(profileids[:10])[:5]:
            print("Profile: %s" % pid)
            profile = data[data.profileid == pid]
            h = HumanQC().eval(
                    profile[varname],
                    profile['PRES'],
                    baseflag=profile['flag_calibrating'],
                    #fails=np.array(profile['mistake']),
                    #fails=ma.array(profile.derr == profile.loc[
                    #    profile.flag_global_range == 1, 'derr'].max()),
                    fails=np.array(profile.derr == profile.derr.max()),
                    humanflag=ma.masked_values(
                        profile['human_flag'], None).astype('object'))

            # Update human_flag only at the new values
            profile.loc[:, 'human_flag'] = h
            for i in profile.index[profile.human_flag.notnull()]:
                data.loc[i, 'human_flag'] = profile.loc[i, 'human_flag']

        data.loc[data.human_flag == 'good', 'flag_calibrating'] = 1
        data.loc[data.human_flag == 'bad', 'flag_calibrating'] = 4
        data.loc[data.human_flag == 'doubt', 'flag_calibrating'] = 0

        result = calibrate4flags(data['flag_calibrating'], data[featuresnames], q=q)


        error_log.append({'err': result['n_err'],
            'err_ratio': result['err_ratio'],
            'p_optimal': result['p_optimal'],
            'tot_misfit': result['tot_misfit']})

        print(error_log[-2])
        print(error_log[-1])

    result['human_flag'] = data['human_flag']
    result['error_log'] = error_log

    return result


def rank_files(datadir, varname, cfg=None):
    """
        Ordered list from datadir files of probably bad data


        The concept is for a recommendation system for Human Q.C.

            - Input: Lista de arquivo | diretório
            - Pandas Collection:
                - Load all files and return flags plus features (aux)
            - Fit features, on full DB, there is no split data
            - With the parameters on the previous step estimate the
                probability of each measurement
            - Create a list sorted by:
                - Produtorium of all probabilities or,
                - Min(P(x_i)), the lowest probability for each measurement
            - The output would be a list to feed the Human Q.C. system

        ATENTION: I must create some tests for this function.
    """
    import pandas as pd

    assert type(varname) is str

    db = ProfilesQCPandasCollection(datadir, cfg=cfg, saveauxiliary=True)

    # hardlimit_flags = ['global_range']
    ind = db.flags[varname]['global_range'] == 1
    features = db.auxiliary[varname][ind]

    params = fit_tests(features, q=.85)
    # Note that I'm already filtering to positions ind, i.e. valid
    #   global range limits. Global range is too obvious and should
    #   be left aside.
    prob = estimate_anomaly(features, params)

    tmp = db.data.loc[ind, ['profilename']]
    tmp.loc[:, 'anomaly_detection'] = pd.Series(prob, index=tmp.index)
    grp = tmp.groupby('profilename')
    output = grp.min().sort_values(by='anomaly_detection').index.tolist()

    return output

    # profilename = np.asanyarray(db.data['profilename'])
    # output = []
    # for pid in np.unique(profilename):
    #     output.append([pid, min(prob[profilename == pid])])

    # return [x[0] for x in sorted(output, key=lambda x: x[1])]



# -*- coding: utf-8 -*-

"""


Shall I use a decorator??

DATA = [25.32, 25.34, 25.34, 25.31, 24.99, 23.46, 21.85, 17.95, 15.39, 11.08, 6.93, 7.93, 5.71, 3.58, np.nan, 1, 1]


tukey53H(np.array, np.maskedArray, pd.Series, xr.DataArray)


    delta = tukey53H(x)

    w = np.hamming(l)
    sigma = (ma.convolve(x, w, mode="same") / w.sum()).std()

    return delta / sigma
"""

import logging

import numpy as np
from numpy import ma

from cotede.qctests import QCCheckVar

try:
    import pandas as pd

    PANDAS_AVAILABLE = True
except:
    PANDAS_AVAILABLE = False


module_logger = logging.getLogger(__name__)


def tukey53H(x, normalize=False):
    """Spike test Tukey 53H from Goring & Nikora 2002

    Return
    ------
    delta :
        An array with the same shape of input x of the difference between x
        and a smoothed x.
    """
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    if not PANDAS_AVAILABLE:
        return _tukey53H_numpy(x, normalize=normalize)

    if hasattr(x, "to_series"):
        x = x.to_series()
    elif not isinstance(x, pd.Series):
        x = pd.Series(x)

    u1 = x.reset_index(drop=True).rolling(5, center=True).median()
    u2 = u1.rolling(3, center=True).median()
    u3 = 0.25 * (u2.shift(-1) + 2 * u2 + u2.shift(1))

    delta = x - u3

    if not normalize:
        return np.array(delta)

    sigma = u1.dropna().std(ddof=1)
    return np.array(delta / sigma)


def _tukey53H_numpy(x, normalize=False):
    """Equivalent to tukey53H but without using pandas

    Note
    ----
    - For larger datasets (>1k) like timeseries the pandas alternative can be
      significantly faster.
    """
    if isinstance(x, ma.MaskedArray):
        x[x.mask] = np.nan
        x = x.data

    N = len(x)

    u1 = np.nan * np.ones(N)
    for n in range(N - 4):
        u1[n + 2] = np.median(x[n : n + 5])

    u2 = np.nan * np.ones(N)
    for n in range(N - 2):
        u2[n + 1] = np.median(u1[n : n + 3])

    u3 = np.nan * np.ones(N)
    u3[1:-1] = 0.25 * (u2[:-2] + 2 * u2[1:-1] + u2[2:])

    delta = np.nan * np.ones(N)
    delta[1:-1] = x[1:-1] - u3[1:-1]

    if not normalize:
        return delta

    idx = ~np.isnan(u1)
    if idx.all():
        return np.nan * delta
    sigma = np.std(u1[idx], ddof=1)
    return delta / sigma


def tukey53H_norm(x, l=12):
    """Spike test Tukey53H() normalized by the std of the low pass

    ATTENTION: l option is temporarily deactivated.

       l is the number of observations. The default l=12 is trully not
         a big number, but this test foccus on spikes, therefore, any
         variability longer than 12 is something else.
    """
    return tukey53H(x, normalize=True)


class Tukey53H(QCCheckVar):
    def set_features(self):
        self.features = {
            "tukey53H": tukey53H(self.data[self.varname]),
            "tukey53H_norm": tukey53H_norm(self.data[self.varname], l=self.cfg["l"]),
        }

    def test(self):
        """

                I slightly modified the Goring & Nikora 2002. It is
                  expected that CTD profiles has a typical depth
                  structure, with a range between surface and bottom.
        """
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            print("Deprecated cfg format. It should contain a threshold item.")
            threshold = self.cfg["k"]

        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = np.absolute(self.features["tukey53H"])
        flag[feature > threshold] = self.flag_bad
        flag[feature <= threshold] = self.flag_good
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["tukey53H"] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Deepest Pressure

   Test from Argo QC manual 2.9.1
"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)


class DeepestPressure(QCCheckVar):
    """Deepest measurable pressure

       Some probes have a maximum operational depth, which can be used as a
       hard threshold. For instance, on Argo, Solo operates up to 2000m while
       Deep Solo goes up to 6000m.
    """
    flag_bad = 3

    def test(self, tol_frac=0.1):
        """Apply test to define flags

           tol_frac is the tolerance fraction of the threshold. Argo uses a
           tolerance of 10%, so that every measurement deeper than the 110%
           of the threshold fails this flag.
        """
        self.flags = {}

        assert "PRES" in self.data.keys(), "Missing pressure record"

        threshold = self.cfg["threshold"]
        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )
        threshold *= (1 + tol_frac)

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = self.data["PRES"]
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags["deepest_pressure"] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

Threshold - |median(v0..v4)| + |sigma(v0..v4)|
y = ma.masked_all_like(x)
yy = np.stack([x[:-4], x[1:-3], x[2:-2], x[3:-1], x[4:]])
y[2:-2] = np.median(yy, axis=0) + yy.std(axis=0)
y = np.stack([x[:-4], x[1:-3], x[2:-2], x[3:-1], x[4:]])
"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar


module_logger = logging.getLogger(__name__)


def spike(x):
    """ Spike
    """
    if isinstance(x, ma.MaskedArray):
        mask = x.mask
        x = x.data
        x[mask] = np.nan

    y = np.nan * x
    y[1:-1] = np.abs(x[1:-1] - (x[:-2] + x[2:]) / 2.0) - np.abs((x[2:] - x[:-2]) / 2.0)
    return y


class Spike(QCCheckVar):
    def set_features(self):
        self.features = {"spike": spike(self.data[self.varname])}

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            module_logger.warning(
                "Deprecated cfg format. It should contain a threshold item."
            )
            threshold = self.cfg

        assert (
            (np.size(threshold) == 1)
            and (threshold is not None)
            and (np.isfinite(threshold))
        )

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = np.absolute(self.features["spike"])
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        # Flag as 9 any masked input value
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["spike"] = flag



# -*- coding: utf-8 -*-

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheck

module_logger = logging.getLogger(__name__)

try:
    import oceansdb
except ImportError:
    module_logger.debug("OceansDB package is not available")


def location_at_sea(data, cfg=None):
    """ Evaluate if location is at sea.

        Interpolate the depth from ETOPO for the data position.
          If local "height" is negative, means lower than sea
          level, hence at sea.

        FIXME: It must allow to check Lat/Lon from data to work with
          TSGs, i.e. one location for each measurement.
          Double check other branches, I thought I had already done
            this before.
    """
    try:
        flag_good = cfg["flag_good"]
    except:
        flag_good = 1
    try:
        flag_bad = cfg["flag_bad"]
    except:
        flag_bad = 3

    assert hasattr(data, "attrs"), "Missing attributes"

    # Temporary solution while migrating to OceanSites variables syntax
    if ("LATITUDE" not in data.attrs) and ("latitude" in data.attrs):
        module_logger.debug(
            "Deprecated. In the future it will not accept latitude anymore. It'll must be LATITUDE"
        )
        data.attrs["LATITUDE"] = data.attrs["latitude"]
    if ("LONGITUDE" not in data.attrs) and ("longitude" in data.attrs):
        module_logger.debug(
            "Deprecated. In the future it will not accept longitude anymore. It'll must be LONGITUDE"
        )
        data.attrs["LONGITUDE"] = data.attrs["longitude"]

    if (
        ("LATITUDE" not in data.attrs)
        or (data.attrs["LATITUDE"] == None)
        or ("LONGITUDE" not in data.attrs)
        or (data.attrs["LONGITUDE"] == None)
    ):
        module_logger.debug("Missing geolocation (lat/lon)")
        return 0

    if (
        (data.attrs["LATITUDE"] > 90)
        or (data.attrs["LATITUDE"] < -90)
        or (data.attrs["LONGITUDE"] > 360)
        or (data.attrs["LONGITUDE"] < -180)
    ):
        return flag_bad

    try:
        ETOPO = oceansdb.ETOPO()
        etopo = ETOPO["topography"].extract(
            var="height", lat=data.attrs["LATITUDE"], lon=data.attrs["LONGITUDE"]
        )
        h = etopo["height"]

        flag = np.zeros(h.shape, dtype="i1")
        flag[np.nonzero(h <= 0)] = flag_good
        flag[np.nonzero(h > 0)] = flag_bad

        return flag

    except:
        return 0


def get_bathymetry(lat, lon, resolution="5min"):
    """Interpolate bathymetry from ETOPO

       For a given (lat, lon), interpolates the bathymetry from ETOPO
    """
    assert np.shape(lat) == np.shape(lon), "Lat & Lon shape mismatch"

    db = oceansdb.ETOPO(resolution=resolution)

    etopo = db["topography"].track(var="height", lat=lat, lon=lon)
    return {"bathymetry": -etopo["height"].astype("i")}


class LocationAtSea(QCCheck):
    def set_features(self):
        # if ("latitude" in self.data.keys()) and ("longitude" in self.data.keys()):
        #     lat = self.data["latitude"]
        #     lon = self.data["longitude"]
        if ("LATITUDE" in self.data.keys()) and ("LONGITUDE" in self.data.keys()):
            lat = self.data["LATITUDE"]
            lon = self.data["LONGITUDE"]
        elif ("LATITUDE" in self.data.attrs) and ("LONGITUDE" in self.data.attrs):
            lat = self.data.attrs["LATITUDE"]
            lon = self.data.attrs["LONGITUDE"]
        else:
            module_logger.debug("Missing geolocation (lat/lon)")
            self.features = {}
            return

        try:
            self.features = get_bathymetry(lat=lat, lon=lon)
            # idx = np.isfinite(lat) & np.isfinite(lon)
            # self.features = get_bathymetry(lat=lat[idx], lon=lon[idx])
        except:
            self.features = {
                "bathymetry": ma.fix_invalid([np.nan]),
                "bathymetry_std": ma.fix_invalid([np.nan]),
            }
        return

        if (
            ("LATITUDE" not in self.data.attrs)
            or (self.data.attrs["LATITUDE"] is None)
            or ("LONGITUDE" not in self.data.attrs)
            or (self.data.attrs["LONGITUDE"] is None)
        ):
            self.features = {
                "bathymetry": ma.fix_invalid([np.nan]),
                "bathymetry_std": ma.fix_invalid([np.nan]),
            }
            self.flags["valid_position"] = self.flag_bad
            return

        if (
            (self.data.attrs["LATITUDE"] > 90)
            or (self.data.attrs["LATITUDE"] < -90)
            or (self.data.attrs["LONGITUDE"] > 360)
            or (self.data.attrs["LONGITUDE"] < -180)
        ):
            self.features = {
                "bathymetry": ma.fix_invalid([np.nan]),
                "bathymetry_std": ma.fix_invalid([np.nan]),
            }
            return

        lat = self.data.attrs["LATITUDE"]
        lon = self.data.attrs["LONGITUDE"]
        try:
            self.features = get_bathymetry(lat=lat, lon=lon)
        except:
            self.features = {
                "bathymetry": ma.fix_invalid([np.nan]),
                "bathymetry_std": ma.fix_invalid([np.nan]),
            }

    def test(self):
        self.flags = {}
        try:
            threshold = self.cfg["threshold"]
        except KeyError:
            module_logger.debug("No threshold for location at sea. I'll use 0")
            threshold = 0

        flag = np.zeros(self.features["bathymetry"].shape, dtype="i1")
        flag[np.nonzero(self.features["bathymetry"] < threshold)] = self.flag_bad
        flag[np.nonzero(self.features["bathymetry"] >= threshold)] = self.flag_good
        flag[ma.getmaskarray(self.features["bathymetry"])] = 9
        self.flags["location_at_sea"] = flag



# -*- coding: utf-8 -*-

""" The tests itself to QC the data
"""

import logging

import numpy as np
from numpy import ma

module_logger = logging.getLogger(__name__)


# FIXME, tests to adjust on public version
def frozen_profile():
    pass


def grey_list():
    pass


def valid_speed():
    pass


def gross_sensor_drift():
    pass


def platform_identification():
    pass


class QCCheck(object):
    """Basic template for a QC check
    """

    flag_good = 1
    flag_bad = 4

    def __init__(self, data, cfg=None, autoflag=True):
        self.data = data
        if (cfg is not None) or not hasattr(self, 'cfg'):
            self.cfg = cfg

        self.set_flags()
        self.set_features()
        if autoflag:
            self.test()

    def __getitem__(self, key):
        return self.data[key]

    @property
    def attrs(self):
        return self.data.attrs

    def set_features(self):
        self.features = {}

    def set_flags(self):
        try:
            self.flag_good = self.cfg["flag_good"]
        except (KeyError, TypeError):
            module_logger.debug("flag_good undefined. Using default value")

        try:
            self.flag_bad = self.cfg["flag_bad"]
        except (KeyError, TypeError):
            module_logger.debug("flag_bad undefined. Using default value")

    def keys(self):
        return self.features.keys() + ["flag_%s" % f for f in self.flags.keys()]


class QCCheckVar(QCCheck):
    """Template for a QC check of a specific variable
    """

    def __init__(self, data, varname, cfg=None, autoflag=True):
        self.varname = varname
        super().__init__(data, cfg, autoflag)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE_scikit-fuzzy.txt

"""

    This was strongly inspired on scikit-fuzzy by Joshua Warner et.al.,
      and they deserve most of the credits for this.
    The decision to split came from the fact that here we use only few
      functions from scikit-fuzzy, and scikit-fuzzy 0.1.9 stopped to
      work with pypi, breaking all tests and the development process of
      CoTeDe.

"""

import numpy as np

def defuzz(x, mfx, mode):
    """
    Defuzzification of a membership function, returning a defuzzified value
    of the function at x, using various defuzzification methods.

    Parameters
    ----------
    x : 1d array or iterable, length N
        Independent variable.
    mfx : 1d array of iterable, length N
        Fuzzy membership function.
    mode : string
        Controls which defuzzification method will be used.
        * 'centroid': Centroid of area
        * 'bisector': bisector of area
        * 'mom'     : mean of maximum
        * 'som'     : min of maximum
        * 'lom'     : max of maximum

    Returns
    -------
    u : float or int
        Defuzzified result.

    """
    mode = mode.lower()
    x = x.ravel()
    mfx = mfx.ravel()
    n = len(x)
    assert n == len(mfx), 'Length of x and fuzzy membership function must be \
                          identical.'

    if 'centroid' in mode or 'bisector' in mode:
        tot_area = mfx.sum()
        assert tot_area != 0, 'Total area is zero in defuzzification!'

        if 'centroid' in mode:
            return centroid(x, mfx)

        elif 'bisector' in mode:
            tmp = 0
            for k in range(n):
                tmp += mfx[k]
                if tmp >= tot_area / 2.:
                    return x[k]

    elif 'mom' in mode:
        return np.mean(x[mfx == mfx.max()])

    elif 'som' in mode:
        tmp = x[mfx == mfx.max()]
        return tmp[tmp == np.abs(tmp).min()][0]

    elif 'lom' in mode:
        tmp = x[mfx == mfx.max()]
        return tmp[tmp == np.abs(tmp).max()][0]

    else:
        raise ValueError('The input for `mode`, %s, was incorrect.' % (mode))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

8. Pressure increasing test
This test requires that the profile has pressures that are monotonically increasing (assuming the pressures are ordered from smallest to largest).
Action: If there is a region of constant pressure, all but the first of a consecutive set of constant pressures should be flagged as bad data. If there is a region where pressure reverses, all of the pressures in the reversed part of the profile should be flagged as bad data. All pressures flagged as bad data and all of the associated temperatures and salinities are removed from the TESAC distributed on the GTS.

"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar


module_logger = logging.getLogger(__name__)


class MonotonicZ(QCCheckVar):
    """Check if sensor vertical movement is monotonic

       This is usually called Increasing Depth, Increasing Pressure ...

       Most of the implementations define a stop or a invertion in the
       verticalm movement as bad data.

       cfg[coord, tolerance]
    """

    coord = "depth"

    def test(self):
        """

           coord = depth
           tolerance = 0.0
        """
        self.flags = {}

        if "coord" in self.cfg:
            self.coord = self.cfg["coord"]

        z = self[self.coord]
        assert np.shape(self[self.varname]) == np.shape(z)

        flag = np.zeros(self[self.varname].shape, dtype="i1")

        dz = np.diff(z)
        if np.all(dz > 0):
            flag[:] = self.flag_good
        else:
            flag[0] = self.flag_good
            zmax = z[0]
            for i, zn in enumerate(z[1:], 1):
                if zn > zmax:
                    zmax = zn
                    flag[i] = self.flag_good
                else:
                    flag[i] = self.flag_bad
        flag[ma.getmaskarray(z)] = 9

        if "flag_name" in self.cfg:
            flag_name = self.cfg["flag_name"]
        else:
            flag_name = "monotonic_{}".format(self.cfg["coord"].lower())
        self.flags[flag_name] = flag



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from numpy import ma

from .qctests import QCCheckVar


def constant_cluster_size(x, tol=0):
    """Estimate the cluster size with (nearly) constant value

       Returns how many consecutive neighbor values are within a given
         tolerance range. Note that invalid values, like NaN, are ignored.
    """
    assert np.ndim(x) == 1, 'Not ready for more than 1 dimension'

    # Adding a tolerance to handle roundings due to different numeric types.
    tol = tol + 1e-5 * tol

    ivalid = np.nonzero(~ma.getmaskarray(ma.fix_invalid(x)))[0]
    dx = np.diff(np.atleast_1d(x)[ivalid])

    cluster_size = np.zeros(np.shape(x), dtype='i')
    for i, iv in enumerate(ivalid):
        idx = np.absolute(dx[i:].cumsum()) > tol
        if True in idx:
            cluster_size[iv] += np.nonzero(idx)[0].min()
        else:
            cluster_size[iv] += idx.size
        idx = np.absolute(dx[0:i][::-1].cumsum()) > tol
        if True in idx:
            cluster_size[iv] += np.nonzero(idx)[0].min()
        else:
            cluster_size[iv] += idx.size
    return cluster_size


class ConstantClusterSize(QCCheckVar):
    """
       Need to implement a check on time. TSG specifies constant value during 6 hrs.
    """
    def set_features(self):
        cluster_size = constant_cluster_size(self.data[self.varname])
        N = ma.compressed(self.data[self.varname]).size
        cluster_fraction = cluster_size / N

        self.features = {'constant_cluster_size': cluster_size,
                         'constant_cluster_fraction': cluster_fraction,
                         }

    def test(self):
        self.flags = {}
        threshold = self.cfg['threshold']

        # assert (np.size(threshold) == 1) \
        #        and (threshold is not None) \
        #        and (np.isfinite(threshold))

        if isinstance(threshold, str) and (threshold[-1] == '%'):
            threshold = float(threshold[:-1]) * 1e-2
            feature_name = 'constant_cluster_fraction'
        else:
            feature_name = 'constant_cluster_size'

        flag = np.zeros(self.data[self.varname].shape, dtype='i1')
        feature = self.features[feature_name]
        flag[np.nonzero(feature > threshold)] = self.flag_bad
        flag[np.nonzero(feature <= threshold)] = self.flag_good
        flag[ma.getmaskarray(self.data[self.varname])] = 9
        self.flags[feature_name] = flag



# -*- coding: utf-8 -*-

"""
"""

from datetime import timedelta
import logging

import numpy as np
from numpy import ma
from oceansdb import CARS

from .qctests import QCCheckVar


module_logger = logging.getLogger(__name__)


class CARS_NormBias(QCCheckVar):
    """Compares measuremnts with CARS climatology

    Notes
    -----
    * Although using standard error is a good idea, the default is to not use
      standard error to estimate the bias to follow the traaditional approach.
      This can have a signifcant impact in the deep oceans and regions lacking
      extensive sampling.
    """

    flag_bad = 3
    use_standard_error = False

    def __init__(self, data, varname, cfg=None, autoflag=True):
        try:
            self.use_standard_error = cfg["use_standard_error"]
        except (KeyError, TypeError):
            module_logger.debug("use_standard_error undefined. Using default value")
        super().__init__(data, varname, cfg, autoflag)

    def keys(self):
        return self.features.keys() + ["flag_%s" % f for f in self.flags.keys()]

    def set_features(self):

        if ("LATITUDE" in self.data.attributes.keys()) and (
            "LONGITUDE" in self.data.attributes.keys()
        ):
            kwargs = {
                "lat": self.data.attributes["LATITUDE"],
                "lon": self.data.attributes["LONGITUDE"],
            }

        if ("LATITUDE" in self.data.keys()) and ("LONGITUDE" in self.data.keys()):
            dLmax = max(
                data["LATITUDE"].max() - data["LATITUDE"].min(),
                data["LONGITUDE"].max() - data["LONGITUDE"].min(),
            )
            # Only use each measurement coordinate if it is spread.
            if dLmax >= 0.01:
                kwargs = {
                    "lat": self.data["LATITUDE"],
                    "lon": self.data["LONGITUDE"],
                    "alongtrack_axis": ["lat", "lon"],
                }

        if "DEPTH" in self.data.keys():
            depth = self.data["DEPTH"]
        elif "PRES" in self.data.keys():
            depth = self.data["PRES"]

        try:
            doy = int(self.data.attributes["date"].strftime("%j"))
        except:
            doy = int(self.data.attributes["datetime"].strftime("%j"))

        db = CARS()
        if self.varname[-1] == "2":
            vtype = self.varname[:-1]
        else:
            vtype = self.varname

        idx = ~ma.getmaskarray(depth) & np.array(depth >= 0)
        cars = db[vtype].extract(
            var=["mn", "std_dev"], doy=doy, depth=depth[idx], **kwargs
        )

        if idx.all() is not True:
            for v in cars.keys():
                tmp = ma.masked_all(depth.shape, dtype=cars[v].dtype)
                tmp[idx] = cars[v]
                cars[v] = tmp

        self.features = {"cars_mean": cars["mn"], "cars_std": cars["std_dev"]}

        self.features["cars_bias"] = (
            self.data[self.varname] - self.features["cars_mean"]
        )

        # if use_standard_error = True, the comparison with the climatology
        #   considers the standard error, i.e. the bias will be only the
        #   ammount above the standard error range.
        assert not self.use_standard_error, "Sorry, I'm not ready for that"
        if self.use_standard_error is True:
            standard_error = (
                self.features["cars_std"] / self.features["cars_nsamples"] ** 0.5
            )
            idx = np.absolute(self.features["cars_bias"]) <= standard_error
            self.features["cars_bias"][idx] = 0
            idx = np.absolute(self.features["cars_bias"]) > standard_error
            self.features["cars_bias"][idx] -= (
                np.sign(self.features["cars_bias"][idx]) * standard_error[idx]
            )

        self.features["cars_normbias"] = (
            self.features["cars_bias"] / self.features["cars_std"]
        )

    def test(self):

        # 3 is the possible minimum to estimate the std, but I shold use higher.
        try:
            min_samples = self.cfg["min_samples"]
        except KeyError:
            min_samples = 3

        self.flags = {}

        threshold = self.cfg["threshold"]
        assert (np.size(threshold) == 1) and (threshold is not None)

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")

        normbias_abs = np.absolute(self.features["cars_normbias"])
        ind = np.nonzero(normbias_abs <= threshold)
        flag[ind] = self.flag_good
        ind = np.nonzero(normbias_abs > threshold)
        flag[ind] = self.flag_bad

        # Flag as 9 any masked input value
        flag[ma.getmaskarray(self.data[self.varname])] = 9

        self.flags["cars_normbias"] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE_scikit-fuzzy.txt

"""

    This was strongly inspired on scikit-fuzzy by Joshua Warner et.al.,
      and they deserve most of the credits for this.
    The decision to split came from the fact that here we use only few
      functions from scikit-fuzzy, and scikit-fuzzy 0.1.9 stopped to
      work with pypi, breaking all tests and the development process of
      CoTeDe.

"""

import numpy as np
from numpy import ma


def smf(x, p):
    """
    S-function fuzzy membership generator.

    Parameters
    ----------
    x : any sequence
        Independent variable.
    p: list of 2 values
        p[0]: 'foot', where the function begins to climb from zero.
        p[1]: 'ceiling', where the function levels off at 1.

    Returns
    -------
    y : 1d array
        S-function.

    Notes
    -----
    Named such because of its S-like shape.

    """
    assert len(p) == 2, 'smf requires 2 parameters'
    assert p[0] <= p[1], 'p[0] must be <= p[1].'

    x = np.asanyarray(x)
    y = np.ones_like(x)

    idx = x <= p[0]
    y[idx] = 0

    idx = np.logical_and(p[0] <= x, x <= (p[0] + p[1]) / 2.)
    y[idx] = 2. * ((x[idx] - p[0]) / (p[1] - p[0])) ** 2.

    idx = np.logical_and((p[0] + p[1]) / 2. <= x, x <= p[1])
    y[idx] = 1 - 2. * ((x[idx] - p[1]) / (p[1] - p[0])) ** 2.

    return y


def trapmf(x, p):
    """
    Trapezoidal membership function generator.

    Parameters
    ----------
    x : any sequence
        Independent variable.
    p: list of 4 values
        lower than p[0] and higher than p[3] it returns 0
        between p[1] and p[2] it returns 1
        linear transition from p[0] to p[1] and from p[2] to p[3]

    Returns
    -------
    y : 1d array
        Trapezoidal membership function.

    """
    assert len(p) == 4, 'trapmf requires 4 parameters.'
    assert p[0] <= p[1] and p[1] <= p[2] and p[2] <= p[3], \
            'trapmf requires 4 parameters: p[0] <= p[1] <= p[2] <= p[3].'

    x = np.asanyarray(x)
    y = np.ones_like(x)

    idx = np.nonzero(x <= p[1])
    y[idx] = trimf(x[idx], [p[0], p[1], p[1]])

    idx = np.nonzero(x >= p[2])[0]
    y[idx] = trimf(x[idx], [p[2], p[2], p[3]])

    idx = np.nonzero(x < p[0])[0]
    y[idx] = np.zeros(len(idx))

    idx = np.nonzero(x > p[3])[0]
    y[idx] = np.zeros(len(idx))

    return y


def trimf(x, p):
    """
    Triangular membership function generator.

    Parameters
    ----------
    x : any sequence
        Independent variable.
    p: list of 4 values
        lower than p[0] and higher than p[3] it returns 0
        between p[1] and p[2] it returns 1

    Returns
    -------
    y : 1d array
        Triangular membership function.

    """
    assert len(p) == 3, 'trimf requires 3 parameters.'
    assert p[0] <= p[1] and p[1] <= p[2], \
            'trimf requires 3 parameters: p[0] <= p[1] <= p[2].'

    x = np.asanyarray(x)
    y = np.zeros(x.shape)

    # Left side
    if p[0] != p[1]:
        idx = np.nonzero(np.logical_and(p[0] < x, x < p[1]))
        y[idx] = (x[idx] - p[0]) / float(p[1] - p[0])

    # Right side
    if p[1] != p[2]:
        idx = np.nonzero(np.logical_and(p[1] < x, x < p[2]))[0]
        y[idx] = (p[2] - x[idx]) / float(p[2] - p[1])

    idx = np.nonzero(x == p[1])
    y[idx] = 1

    return y


def zmf(x, p):
    """
    Z-function fuzzy membership generator.

    Parameters
    ----------
    x : any sequence
        Independent variable.
    p: list of 2 values
        p[0]: 'foot', where the function begins to climb from zero.
        p[1]: 'ceiling', where the function levels off at 1.

    Returns
    -------
    y : 1d array
        Z-function.

    Notes
    -----
    Named such because of its Z-like shape.

    """
    assert len(p) == 2, 'zmf requires 2 parameters'
    assert p[0] <= p[1], 'p[0] must be <= p[1].'

    x = np.asanyarray(x)
    y = np.ones_like(x)

    idx = np.logical_and(p[0] <= x, x < (p[0] + p[1]) / 2.)
    y[idx] = 1 - 2. * ((x[idx] - p[0]) / (p[1] - p[0])) ** 2.

    idx = np.logical_and((p[0] + p[1]) / 2. <= x, x <= p[1])
    y[idx] = 2. * ((x[idx] - p[1]) / (p[1] - p[0])) ** 2.

    idx = x >= p[1]
    y[idx] = 0

    return y



#!/usr/bin/env python

from .qctests import *
from .descentPrate import descentPrate
from .anomaly_detection import anomaly_detection
from .possible_speed import possible_speed
from .morello2014 import morello2014
from .fuzzylogic import fuzzylogic

from .bin_spike import Bin_Spike, bin_spike
from .cars_normbias import CARS_NormBias
from .constant_cluster_size import ConstantClusterSize, constant_cluster_size
from .cum_rate_of_change import CumRateOfChange, cum_rate_of_change
from .deepest_pressure import DeepestPressure
from .density_inversion import DensityInversion, densitystep
from .digit_roll_over import DigitRollOver
from .global_range import GlobalRange
from .gradient import Gradient, gradient
from .gradient_depthconditional import GradientDepthConditional
from .location_at_sea import LocationAtSea, location_at_sea
from .monotonic_z import MonotonicZ
from .profile_envelop import ProfileEnvelop, profile_envelop
from .regional_range import RegionalRange
from .rate_of_change import RateOfChange, rate_of_change
from .spike import Spike, spike
from .spike_depthconditional import SpikeDepthConditional
from .tukey53H import Tukey53H, tukey53H, tukey53H_norm
from .woa_normbias import WOA_NormBias, woa_normbias
from .stuck_value import StuckValue
from .valid_geolocation import ValidGeolocation



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst


import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read()
    if sys.version_info[0] == 2:
        requirements.replace('gsw>=3.0.6', 'gsw==3.0.6')


setup(
    name='cotede',
    version='0.23.2',
    description='Quality Control of Oceanographic Data',
    long_description=readme + '\n\n' + history,
    author='Guilherme Castelão',
    author_email='guilherme@castelao.net',
    url='http://cotede.castelao.net',
    packages=[
        'cotede',
        'cotede.qctests',
        'cotede.utils',
        'cotede.humanqc',
        'cotede.anomaly_detection',
        'cotede.fuzzy',
    ],
    package_dir = {'cotede':
                   'cotede'},
    license='3-clause BSD',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: BSD License',
        ],
    keywords='CTD TSG SeaBird ARGO Quality Control oceanography hydrography',
    include_package_data=True,
    zip_safe=False,
    extras_require = {
        'GSW': ["gsw>=3.0.6"],
        'OceansDB': ["oceansdb>=0.8.6"],
        'manualqc': ["matplotlib"],
        'regional': ["Shapely>=1.6.4"]
    }
)



# -*- coding: utf-8 -*-

import numpy as np
from numpy import ma
from numpy import radians

AVG_EARTH_RADIUS = 6371000  # in m


def haversine(lat_a, lon_a, lat_b, lon_b):
    """

        Copied from MAUD.
    """

    lat_a, lon_a, lat_b, lon_b = map(radians, [lat_a, lon_a, lat_b, lon_b])

    dlat = lat_a - lat_b
    dlon = lon_a - lon_b
    d = np.sin(dlat / 2) ** 2 + np.cos(lat_a) * np.cos(lat_b) * np.sin(dlon / 2) ** 2
    h = 2 * AVG_EARTH_RADIUS * np.arcsin(np.sqrt(d))

    return h


def speed(data):
    """
    """
    assert "timeS" in data.keys(), "Missing timeS in input data"
    assert "LATITUDE" in data.keys(), "Missing LATITUDE in input data"
    assert "LONGITUDE" in data.keys(), "Missing LONGITUDE in input data"

    dL = haversine(
        data["LATITUDE"][:-1],
        data["LONGITUDE"][:-1],
        data["LATITUDE"][1:],
        data["LONGITUDE"][1:],
    )
    dt = ma.diff(data["timeS"])

    speed = ma.append(ma.masked_array([0], [True]), dL / dt)

    return speed


def possible_speed(data, cfg):
    """

        Consider 4 if smooth is higher, 3 if smooth smaller, 1 if raw smaller
    """
    s = speed(data)

    flag = np.zeros(s.shape, dtype="i1")

    flag[s <= cfg["threshold"]] = cfg["flag_good"]
    flag[s > cfg["threshold"]] = cfg["flag_bad"]

    # Flag as 9 any masked input value
    # for v in ['LATITUDE', 'LONGITUDE']:
    #    flag[ma.getmaskarray(data[v])] = 9
    # I'm not sure if it's a good idea. I should flag as 9 if the variable
    #   being tested is invalid. For example, if I have a good temperature,
    #   but no valid location, the possible_speed test should return 0, i.e.
    #   no Q.C. evaluated.

    return flag, s



# -*- coding: utf-8 -*-

"""
    Quality Control based on fuzzy logic.
"""

import numpy as np
from cotede.fuzzy import fuzzy_uncertainty


def fuzzylogic(features, cfg):
    """

        FIXME: Think about, should I return 0, or have an assert, and at qc.py
          all qc tests are applied with a try, and in case it fails it flag
          0s.

    """
    # FIXME: Should I use all or any? Should I still try to apply fuzzy even
    #   when it is missing some of the features? Sounds like the best guess
    #   possible.
    if not np.all([f in features for f in cfg['features']]):
        print("Not all features (%s) required to fuzzyfy are available" %
                cfg['features'].keys())
        try:
            return np.zeros(features[features.keys()[0]].shape, dtype='i1')
        except:
            return 0

    #f = fuzzyfy(features, cfg)
    uncertainty = fuzzy_uncertainty(features, cfg)

    # FIXME: As it is now, it will have no zero flag value. Think about cases
    #   where some values in a profile would not be estimated, hence flag=0
    # I needed to use np.nonzeros because now uncertainty is a masked array,
    #   to accept when a feature is masked.
    flags = np.zeros(
            features[list(cfg['features'].keys())[0]].shape, dtype='i1')
    flags[np.nonzero(uncertainty <= 0.29)] = 1
    flags[np.nonzero((uncertainty > 0.29)  & (uncertainty <= 0.34))] = 2
    flags[np.nonzero((uncertainty > 0.34)  & (uncertainty <= 0.72))] = 3
    flags[np.nonzero(uncertainty > 0.72)] = 4

    return flags



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar
from .gradient import curvature

module_logger = logging.getLogger(__name__)


class GradientDepthConditional(QCCheckVar):
    def set_features(self):
        self.features = {"gradient": curvature(self.data[self.varname])}

    def test(self):
        self.flags = {}

        flag = np.zeros(self.data[self.varname].shape, dtype="i1")
        feature = np.absolute(self.features["gradient"])

        # ---- Shallow zone -----------------
        threshold = self.cfg["shallow_max"]
        flag[np.nonzero(
                (self["PRES"] <= self.cfg["pressure_threshold"])
                & (feature > threshold)
            )
        ] = self.flag_bad
        flag[
            np.nonzero(
                (self["PRES"] <= self.cfg["pressure_threshold"])
                & (feature <= threshold)
            )
        ] = self.flag_good
        # ---- Deep zone --------------------
        threshold = self.cfg["deep_max"]
        flag[
            np.nonzero(
                (self["PRES"] > self.cfg["pressure_threshold"])
                & (feature > threshold)
            )
        ] = self.flag_bad
        flag[
            np.nonzero(
                (self["PRES"] > self.cfg["pressure_threshold"])
                & (feature <= threshold)
            )
        ] = self.flag_good

        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["gradient_depthconditional"] = flag



# -*- coding: utf-8 -*-

"""
    Quality Control based on anomaly detection
"""


import numpy as np
from numpy import ma
from ..fuzzy import fuzzyfy


def morello2014(features, cfg):
    """

        FIXME: Think about, should I return 0, or have an assert, and at qc.py
          all qc tests are applied with a try, and in case it fails it flag
          0s.

    """
    # for f in cfg['features']:
    #    assert f in features, \
    #            "morello2014 requires feature %s, which is not available" \
    #            % f

    if not np.all([f in features for f in cfg["features"]]):
        print(
            "Not all features (%s) required by morello2014 are available"
            % cfg["features"].keys()
        )
        try:
            return np.zeros(features[features.keys()[0]].shape, dtype="i1")
        except:
            return 0

    f = fuzzyfy(features, cfg)

    for level in f:
        if isinstance(f[level], ma.MaskedArray):
            mask = f[level].mask
            f[level] = f[level].data
            f[level][mask] = np.nan

    # This is how Timms and Morello defined the Fuzzy Logic approach
    # flag = np.zeros(N, dtype='i1')
    # Flag must be np.array, not a ma.array.
    flag = np.zeros(features[list(features.keys())[0]].shape, dtype="i1")

    flag[(f["low"] > 0.5) & (f["high"] < 0.3)] = 2
    flag[(f["low"] > 0.9)] = 1
    # Everything else is flagged 3
    flag[(f["low"] <= 0.5) | (f["high"] >= 0.3)] = 3
    # Missing check if threshold was crossed, to flag as 4
    # The thresholds coincide with the end of the ramp for the fuzzy set high,
    #   hence we can simply
    flag[(f["high"] == 1.0)] = 4

    return flag



# -*- coding: utf-8 -*-

"""Resources related to QC configuration
"""

from collections import OrderedDict
import copy
import json
import logging
import os.path
import pkg_resources

from .utils import cotederc

module_logger = logging.getLogger(__name__)


def list_cfgs():
    """List the available QC procedures, builtin + local

    Full QC procedures, defining which tests and respective parameters to be
    used, can be saved to be re-used later. Several procedures are built-in
    CoTeDe, but the user can create its own collection. This function returns
    a list of all procedures available, built-in + the local user collection.

    See also
    --------
    utils.load_cfg
    """
    cfg = pkg_resources.resource_listdir("cotede", "qc_cfg")
    cfg = sorted([c[:-5] for c in cfg if c[-5:] == ".json"])

    ucfg = os.listdir(cotederc("cfg"))
    ucfg = [c[:-5] for c in ucfg if c[-5:] == ".json"]
    ucfg = [c for c in ucfg if c not in cfg]

    cfg.extend(sorted(ucfg))

    return cfg


def inheritance(child, parent):
    """Aggregate into child what is missing from parent
    """
    for v in child:
        if (v in parent) and isinstance(child[v], dict) and isinstance(parent[v], dict):
            parent[v] = inheritance(child[v], parent[v])
        else:
            parent[v] = child[v]
    return parent


def load_cfg(cfgname="cotede"):
    """Load a QC configuration

    A QC procedure is a sequence of tests, and respective tuning parameters
    used to quality control a dataset. This is how the user controls the
    QC steps to apply.

    Parameters
    ----------
    cfgname : string or dict-list, optional

        - None: If not given, it will use the CoTeDe's default configuration,
          which is equivalent to use cfgname='cotede'.

        - A config name [string]: A string with the name of a json file
          describing the QC procedure. It will first search among the build
          in pre-set (ex.: cotede, eurogoos, gtspp, argo, ...). If can't find
          a config with that name, it will search at ~/.config/cotederc/cfg/,
          or the path defined by the local variable $COTEDE_DIR.

        - Inline config [dict-like]: A dictionary describing the variables to
          process and which tests to use on each one. A minimalist example to
          apply the gradient test in the sea water temperature could be:

              >>> {"sea_water_temperature": {"gradient": 3}}

        If inherit is used, it should be a string or a list of other
        procedures to inherit, where each item has higher priority than the
        following ones. For example:

        >>> {"inherit": "eurogoos", "sea_water_temperature": {"gradient": 2}}

        will use all the setup from eurogoos, and include/overwrite the
        gradient test for sea_water_temperature with a threshold of 2.

    Returns
    -------
    cfg : OrderedDict
        A dictionary defining a full QC procedure that defines which tests to
        run on which variables.

    See also
    --------
    utils.list_cfgs
    """
    if cfgname is None:
        cfgname = "cotede"

    assert isinstance(
        cfgname, (dict, str)
    ), "load_cfg() input must be a dictionary or a str"

    # A given manual configuration has priority
    if isinstance(cfgname, dict):
        module_logger.debug("User's QC cfg: %s" % cfgname)
        cfg = OrderedDict(copy.deepcopy(cfgname))
    elif isinstance(cfgname, str):
        try:
            # If cfg is available in builtin options, use it
            p = pkg_resources.resource_string(
                "cotede", os.path.join("qc_cfg", "{}.json".format(cfgname))
            )
            cfg = json.loads(p, object_pairs_hook=OrderedDict)
            module_logger.debug("Builtin config - %s" % cfgname)
        except:
            # Otherwise, try to load from user's directory
            p = os.path.join(cotederc("cfg"), "{}.json".format(cfgname))
            with open(p, "r") as f:
                cfg = json.load(f, object_pairs_hook=OrderedDict)
            module_logger.debug("User collection cfg - %s" % cfgname)

    cfg = fix_config(cfg)
    if "inherit" in cfg:
        if isinstance(cfg["inherit"], str):
            cfg["inherit"] = [cfg["inherit"]]
        for parent in cfg["inherit"]:
            cfg = inheritance(cfg, load_cfg(parent))

    cfg = fix_procedure(cfg)

    return cfg


def fix_config(cfg):
    """Adjust the config to the latest standard, if necessary

       This function allows backward compatibility with old config descriptos
       updating them to the current standard.
    """
    if ("revision" in cfg) and (cfg["revision"] == "0.21"):
        return cfg

    if "revision" not in cfg:
        cfg = convert_pre_to_021(cfg)

    return cfg


def convert_pre_to_021(cfg):
    """Convert config standard 0.20 into 0.21

       Revision 0.20 is the original standard, which lacked a revision.

       Variables moved from top level to inside item 'variables'.
       Ocean Sites nomenclature moved to CF standard vocabulary:
         - TEMP -> sea_water_temperature
         - PSAL -> sea_water_salinity
    """

    def label(v):
        """Convert Ocean Sites vocabulary to CF standard names
        """
        if v == "PRES":
            return "sea_water_pressure"
        if v == "TEMP":
            return "sea_water_temperature"
        elif v == "PSAL":
            return "sea_water_salinity"
        else:
            return v

    keys = list(cfg.keys())

    output = OrderedDict()
    output["revision"] = "0.21"

    if "inherit" in keys:
        output["inherit"] = cfg["inherit"]
        keys.remove("inherit")

    if "main" in cfg:
        output["common"] = cfg["main"]
        keys.remove("main")
    elif "common" in cfg:
        output["common"] = cfg["common"]
        keys.remove("common")

    def fix_threshold(cfg):
        """Explicit threshold"""
        for t in cfg:
            if isinstance(cfg[t], (int, float)):
                cfg[t] = {"threshold": cfg[t]}
        return cfg

    def fix_regional_range(cfg):
        """Explicit regions
        """
        if "regional_range" in cfg:
            cfg["regional_range"] = {"regions": cfg["regional_range"]}
        return cfg

    def fix_profile_envelop(cfg):
        """Explicit layers

        Note
        ----
        Should I confirm that cfg['profile_envelop'] is a list?
        """
        if ("profile_envelop" in cfg) and ("layers" not in cfg["profile_envelop"]):
            cfg["profile_envelop"] = {"layers": cfg["profile_envelop"]}
        return cfg

    output["variables"] = OrderedDict()
    for k in keys:
        cfg[k] = fix_threshold(cfg[k])
        cfg[k] = fix_regional_range(cfg[k])
        cfg[k] = fix_profile_envelop(cfg[k])
        output["variables"][label(k)] = cfg[k]

    return output


def guess_procedure(name, cfg=None):
    catalog = {
        "anomaly_detection": None,
        "bin_spike": "Bin_Spike",
        "cars_normbias": "CARS_NormBias",
        "constant_cluster_size": "ConstantClusterSize",
        "cum_rate_of_change": "CumRateOfChange",
        "deepest_pressure": "DeepestPressure",
        "density_inversion": "DensityInversion",
        "digit_roll_over": "DigitRollOver",
        "fuzzylogic": None,
        "frozen_profile": None,
        "global_range": "GlobalRange",
        "gradient": "Gradient",
        "gradient_depthconditional": "GradientDepthConditional",
        "grey_list": None,
        "gross_sensor_drift": None,
        "monotonic_z": "MonotonicZ",
        "morello2014": None,
        "platform_identification": None,
        "pressure_increasing": None,
        "profile_envelop": "ProfileEnvelop",
        "pstep": None,
        "rate_of_change": "RateOfChange",
        "regional_range": "RegionalRange",
        "spike": "Spike",
        "spike_depthconditional": "SpikeDepthConditional",
        "stuck_value": "StuckValue",
        "tukey53H": "Tukey53H",
        "tukey53H_norm": "Tukey53H",
        "valid_speed": None,
        "woa_normbias": "WOA_NormBias",
        # 'valid_geolocation': ValidGeolocation,
        "valid_geolocation": None,
    }

    if name in catalog:
        return catalog[name]


def fix_procedure(cfg):
    """If a test procedure is not defined, guess it from the name

    For each variable it is defined a sequence of tests to apply. And each
    test is based in one procedure, but multiple tests could be based on the
    same procedure.

    For instance, one could apply the GlobalRange with two different
    thresholds, a lower one for flag 3 and a higher one for flag 4.
    """

    def fix_each_one(cfg):
        for c in cfg:
            procedure = guess_procedure(c, cfg[c])
            if procedure is not None:
                if cfg[c] is None:
                    cfg[c] = {"procedure": procedure}
                elif "procedure" not in cfg[c]:
                    cfg[c]["procedure"] = procedure
        return cfg

    if "common" in cfg:
        cfg["common"] = fix_each_one(cfg["common"])

    if "variables" in cfg:
        for v in cfg["variables"]:
            cfg["variables"][v] = fix_each_one(cfg["variables"][v])

    return cfg



# -*- coding: utf-8 -*-

"""Utilities for CoTeDe

Miscelaneous resources to support CoTeDe.
"""

import os
from os.path import expanduser
import re
import pkg_resources
import json


def cotederc(subdir=None):
    """Directory with custom configuration for CoTeDe

    To keep the local environment tight, CoTeDe expects to find all local files,
    like pre-set QC procedures, in one single place. This function returns the
    path to that directory.

    Parameters
    ----------
    subdir : str, optional
        Sub-directory inside the base custom directory.

    Returns
    -------
    str
        A path to the local custom files.

        The default path is a directory at the user's home like::

            ~/.config/cotederc/

    Note
    ----
    That default path can be modified by defining the environment variable
    COTEDE_DIR. On bash that could be done like::

        export COTEDE_DIR='/my/other/awesome/path/'

    Note
    ----
    For windows users the path is automatically adjusted to reflect its
    syntax.

    Example
    -------
    A sub-directory for configuration files, named 'cfg', can be determined by::

    >>> cotederc('cfg')
    """
    path = os.getenv("COTEDE_DIR", os.path.join("~", ".config", "cotederc"))
    path = os.path.expanduser(path)
    if subdir is not None:
        path = os.path.join(path, subdir)
    return path


# ============================================================================
def savePQCCollection_pandas(db, filename):
    """ Save

        To Do:
            - Save the files in a tmp file
            - As it saves, creates a md5 of each file
            - Put everything together in a tar.bz2, including the md5 list
            - Delete the tmp file
    """
    import os
    import tempfile
    import tarfile
    import shutil
    import hashlib

    # tar = tarfile.open("%s.tar.bz2" % filename, "w:bz2")
    tar = tarfile.open(filename, "w:bz2")
    tmpdir = tempfile.mkdtemp()

    try:
        # Data
        f = "%s/data.hdf" % (tmpdir)
        db.data.to_hdf(f, "df")
        tar.add(f, arcname="data.hdf")
        # hashlib.md5(open(f, 'rb').read()).digest()
        # hashlib.sha256(open(f, 'rb').read()).digest()
        # Flags
        p = os.path.join(tmpdir, "flags")
        os.mkdir(p)
        for k in db.flags.keys():
            f = os.path.join(p, "flags_%s.hdf" % k)
            db.flags[k].to_hdf(f, "df")
            tar.add(f, arcname="flags/flags_%s.hdf" % k)
        if hasattr(db, "auxiliary"):
            p = os.path.join(tmpdir, "aux")
            os.mkdir(p)
            for k in db.auxiliary.keys():
                f = os.path.join(p, "aux_%s.hdf" % k)
                db.auxiliary[k].to_hdf(f, "df")
                tar.add(f, arcname="aux/aux_%s.hdf" % k)
        tar.close()
    except:
        shutil.rmtree(tmpdir)
        raise
        print("Problems saving the data")
        shutil.rmtree("%s.tar.bz2" % filename)
    finally:
        shutil.rmtree(tmpdir)


def loadPQCCollection_pandas(filename):
    import os
    import tempfile
    import tarfile
    import shutil

    tmpdir = tempfile.mkdtemp()
    tar = tarfile.open(filename, "r:*")
    tar.extractall(path=tmpdir)
    shutil.rmtree(tmpdir)



# -*- coding: utf-8 -*-

"""

    Based on Argo's density inversion test. Test 14 @ Argo QC 2.9.1
"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheck

module_logger = logging.getLogger(__name__)

try:
    import gsw

    nogsw = False
except ImportError:
    module_logger.info("Missing package GSW, used to estimate density when needed.")
    nogsw = True


def densitystep(S, T, P, auto_rotate=False):
    """Estimates the potential density step of successive mesurements

       Expects the data to be recorded along the time, i.e. first measurement
         was recorded first. This makes difference since the first measurement
         has no reference to define the delta change.
       This is relevant for the type of instrument. For instance: XBTs are
         always measured surface to bottom, CTDs are expected the same, but
         Spray underwater gliders measure bottom to surface.
    """
    assert T.shape == P.shape
    assert T.shape == S.shape
    assert T.ndim == 1, "Not ready to densitystep an array ndim > 1"

    if nogsw:
        print("Package gsw is required and is not available.")

    rho0 = gsw.pot_rho_t_exact(S, T, P, 0)
    ds = ma.concatenate([ma.masked_all(1), np.sign(np.diff(P)) * np.diff(rho0)])
    return ma.fix_invalid(ds)


class DensityInversion(QCCheck):
    def __init__(self, data, cfg, autoflag=True):
        assert "TEMP" in data.keys(), "Missing TEMP"
        assert "PSAL" in data.keys(), "Missing PSAL"
        assert "PRES" in data.keys(), "Missing PRES"

        super().__init__(data=data, cfg=cfg, autoflag=autoflag)

    def set_features(self):
        self.features = {
            "densitystep": densitystep(
                self.data["PSAL"], self.data["TEMP"], self.data["PRES"]
            )
        }

    def test(self):
        self.flags = {}

        assert (
            (np.size(self.cfg["threshold"]) == 1)
            and (self.cfg["threshold"] is not None)
            and (np.isfinite(self.cfg["threshold"]))
        )

        flag = np.zeros(self.data["TEMP"].shape, dtype="i1")
        threshold = self.cfg["threshold"]
        feature = self.features["densitystep"]
        flag[np.nonzero(feature < threshold)] = self.flag_bad
        flag[np.nonzero(feature >= threshold)] = self.flag_good
        mask = np.any(
            [ma.getmaskarray(self.data[v]) for v in ["PRES", "TEMP", "PSAL"]], axis=0
        )
        flag[mask] = 9
        self.flags["density_inversion"] = flag



# -*- coding: utf-8 -*-

"""


    Create a test to check if cfg['vars'] does exist in the climatology file,
      to avoid error like have t_mn in one and t_an in the other.

    An idea to improve the climatology test. Gridpoints estimated from few
      measurements should be less trustable. Here I'm using a threshold of at
      least 3 samples to be considered. Deep ocean is in general quite
      stable, so few measurements should be sufficient. One possibility is to
      estimate the standard error, which depends on the ammount of samples and
      estimated standard deviation. This would be the 'uncertainty on the
      estimated average climatology. Any measurement in that range would be
      considered statistically identical to the climatology. Above that
      difference, it would be normalized by the standard deviation. Therefore,
      stable areas would be less tolerant to variability, even with few samples.
"""

from datetime import timedelta
import logging

import numpy as np
from numpy import ma
from oceansdb import WOA

from .qctests import QCCheckVar

module_logger = logging.getLogger(__name__)


def woa_normbias(data, v, cfg):
    """

        FIXME: Move this procedure into a class to conform with the new system
          and include a limit in minimum ammount of samples to trust it. For
          example, consider as masked all climatologic values estimated from
          less than 5 samples.
    """

    # 3 is the possible minimum to estimate the std, but I shold use higher.
    min_samples = 3
    woa = None

    db = WOA()
    if v not in db.keys():
        vtype = v[:-1]
    else:
        vtype = v

    # Temporary solution while I'm not ready to handle tracks.
    if ('LATITUDE' in data) and ('LONGITUDE' in data) \
            and ('LATITUDE' not in data.attributes) \
            and ('LONGITUDE' not in data.attributes):
        if 'datetime' in data.keys():
            d = data['datetime']
        elif ('datetime' in data.attributes):
            d0 = data.attributes['datetime']
            if ('timeS' in data.keys()):
                d = [d0 + timedelta(seconds=s) for s in data['timeS']]
            else:
                d = [data.attributes['datetime']]*len(data['LATITUDE']),

        #woa = woa_track_from_file(
        #        d,
        #        data['LATITUDE'],
        #        data['LONGITUDE'],
        #        cfg['file'],
        #        varnames=cfg['vars'])

        module_logger.error("Sorry, I'm temporary not ready to handle tracks.")
        #woa = db[vtype].get_track(var=['mean', 'standard_deviation'],
        #        doy=d,
        #        depth=[0],
        #        lat=data['LATITUDE'],
        #        lon=data['LONGITUDE'])

    elif ('LATITUDE' in data.attributes.keys()) and \
            ('LONGITUDE' in data.attributes.keys()) and \
            ('PRES' in data.keys()):

                woa = db[vtype].track(
                        var=['mean', 'standard_deviation',
                            'number_of_observations'],
                        doy=int(data.attributes['datetime'].strftime('%j')),
                        depth=data['PRES'],
                        lat=data.attributes['LATITUDE'],
                        lon=data.attributes['LONGITUDE'])

    flag = np.zeros(data[v].shape, dtype='i1')
    features = {}

    try:
        woa_bias = data[v] - woa['mean']
        woa_normbias = woa_bias / woa['standard_deviation']

        ind = np.nonzero((woa['number_of_observations'] >= min_samples) &
                (np.absolute(woa_normbias) <= cfg['sigma_threshold']))
        flag[ind] = 1   # cfg['flag_good']
        ind = np.nonzero((woa['number_of_observations'] >= min_samples) &
                (np.absolute(woa_normbias) > cfg['sigma_threshold']))
        flag[ind] = 3   # cfg['flag_bad']

        # Flag as 9 any masked input value
        flag[ma.getmaskarray(data[v])] = 9

        features = {'woa_bias': woa_bias, 'woa_normbias': woa_normbias,
                'woa_std': woa['standard_deviation'],
                'woa_nsamples': woa['number_of_observations'],
                'woa_mean': woa['mean']}

    finally:
        # self.logger.warnning("%s - WOA is not available at this site" %
        # self.name)
        return flag, features


class WOA_NormBias(QCCheckVar):
    """Compares measurements with WOA climatology

    Notes
    -----
    * Although using standard error is a good idea, the default is to not use
      standard error to estimate the bias to follow the traaditional approach.
      This can have a signifcant impact in the deep oceans and regions lacking
      extensive sampling.
    """

    flag_bad = 3
    use_standard_error = False

    def __init__(self, data, varname, cfg=None, autoflag=True):
        try:
            self.use_standard_error = cfg["use_standard_error"]
        except (KeyError, TypeError):
            module_logger.debug("use_standard_error undefined. Using default value")
        super().__init__(data, varname, cfg, autoflag)


    def set_features(self):
        try:
            doy = int(self.data.attrs['date'].strftime('%j'))
        except:
            doy = int(self.data.attrs['datetime'].strftime('%j'))

        if ('LATITUDE' in self.data.attrs.keys()) and \
                ('LONGITUDE' in self.data.attrs.keys()):
                    mode = 'profile'
                    kwargs = {
                            'lat': self.data.attrs['LATITUDE'],
                            'lon': self.data.attrs['LONGITUDE']}

        if ('LATITUDE' in self.data.keys()) and \
                ('LONGITUDE' in self.data.keys()):
                    mode = 'track'
                    dLmax = max(
                            self.data['LATITUDE'].max() - self.data['LATITUDE'].min(),
                            self.data['LONGITUDE'].max() - self.data['LONGITUDE'].min())
                    # Only use each measurement coordinate if it is spread.
                    if dLmax >= 0.01:
                        kwargs = {
                            'lat': self.data['LATITUDE'],
                            'lon': self.data['LONGITUDE']}

        if ('DEPTH' in self.data.keys()):
            depth = self.data['DEPTH']
        elif ('PRES' in self.data.keys()):
            depth = self.data['PRES']

        db = WOA()
        if self.varname[-1] == '2':
            vtype = self.varname[:-1]
        else:
            vtype = self.varname

        woa_vars = ['mean', 'standard_deviation', 'standard_error',
                    'number_of_observations']

        idx = ~ma.getmaskarray(depth) & np.array(depth >= 0)
        if idx.any():
            if mode == 'track':
                woa = db[vtype].track(
                        var=woa_vars,
                        doy=doy,
                        depth=depth[idx],
                        **kwargs)
            else:
                woa = db[vtype].extract(
                        var=woa_vars,
                        doy=doy,
                        depth=depth[idx],
                        **kwargs)
        else:
            woa = {v: ma.masked_all(1) for v in woa_vars}

        if idx.all() is not True:
            for v in woa.keys():
                tmp = ma.masked_all(depth.shape, dtype=woa[v].dtype)
                tmp[idx] = woa[v]
                woa[v] = tmp

        self.features = {
                'woa_mean': woa['mean'],
                'woa_std': woa['standard_deviation'],
                'woa_nsamples': woa['number_of_observations'],
                'woa_se': woa['standard_error']}

        self.features['woa_bias'] = self.data[self.varname] - \
                self.features['woa_mean']

        # if use_standard_error = True, the comparison with the climatology
        #   considers the standard error, i.e. the bias will be only the
        #   ammount above the standard error range.
        if self.use_standard_error is True:
            standard_error = self.features['woa_std'] / \
                    self.features['woa_nsamples'] ** 0.5
            idx = np.absolute(self.features['woa_bias']) <= \
                    standard_error
            self.features['woa_bias'][idx] = 0
            idx = np.absolute(self.features['woa_bias']) > standard_error
            self.features['woa_bias'][idx] -= \
                    np.sign(self.features['woa_bias'][idx]) * \
                    standard_error[idx]

        self.features['woa_normbias'] = self.features['woa_bias'] / \
                self.features['woa_std']

    def test(self):

        # 3 is the possible minimum to estimate the std, but I shold use higher.
        try:
            min_samples = self.cfg['min_samples']
        except KeyError:
            min_samples = 3

        self.flags = {}

        threshold = self.cfg['threshold']
        assert (np.size(threshold) == 1) and \
                (threshold is not None)

        flag = np.zeros(self.data[self.varname].shape, dtype='i1')

        normbias_abs = np.absolute(self.features['woa_normbias'])
        ind = np.nonzero((self.features['woa_nsamples'] >= min_samples) &
                (normbias_abs <= threshold))
        flag[ind] = self.flag_good
        ind = np.nonzero((self.features['woa_nsamples'] >= min_samples) &
                (normbias_abs > threshold))
        flag[ind] = self.flag_bad

        # Flag as 9 any masked input value
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9

        self.flags['woa_normbias'] = flag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""

    ATENTION: Create test for global_range
    - masked values
    - input np.array
    - properly handle inf, nan?
"""

import logging

import numpy as np
from numpy import ma

from .qctests import QCCheckVar


module_logger = logging.getLogger(__name__)


class GlobalRange(QCCheckVar):
    def test(self):
        self.flags = {}
        assert ("minval" in self.cfg) and (
            "maxval" in self.cfg
        ), "Missing limits: minval & maxval"

        assert self.cfg["minval"] < self.cfg["maxval"], (
            "Global Range(%s): "
            + "minval (%s) must be smaller than maxval(%s)"
            % (v, self.cfg["minval"], self.cfg["maxval"])
        )

        minval = self.cfg["minval"]
        maxval = self.cfg["maxval"]

        feature = self.data[self.varname]
        if isinstance(feature, ma.MaskedArray):
            feature[feature.mask] = np.nan
            feature = feature.data

        flag = np.zeros(feature.shape, dtype="i1")
        flag[np.nonzero(feature < minval)] = self.flag_bad
        flag[np.nonzero(feature > maxval)] = self.flag_bad
        idx = (feature >= minval) & (feature <= maxval)
        flag[np.nonzero(idx)] = self.flag_good
        x = self.data[self.varname]
        flag[ma.getmaskarray(x) | ~np.isfinite(x)] = 9
        self.flags["global_range"] = flag
